import scipy.constants as codata
angstroms_to_eV = codata.h*codata.c/codata.e*1e10

from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
from wofry.propagator.propagator import Propagator2D

from wofrysrw.propagator.wavefront2D.srw_wavefront import WavefrontPropagationParameters, WavefrontPropagationOptionalParameters
from wofrysrw.propagator.wavefront2D.srw_wavefront import SRWWavefront

from wofrysrw.util.srw import *

class FresnelSRWWofry(Propagator2D):

    HANDLER_NAME = "FRESNEL_SRW_WOFRY"

    def get_handler_name(self):
        return self.HANDLER_NAME

    """
    2D Fresnel propagator using convolution via Fourier transform
    :param wavefront:
    :param propagation_distance:
    :param srw_autosetting:set to 1 for automatic SRW redimensionate wavefront
    :return:
    """


    def do_specific_progation_before(self, wavefront, propagation_distance, parameters, element_index=None):
        return self.do_specific_progation(wavefront, propagation_distance, parameters, prefix="before")

    def do_specific_progation_after(self, wavefront, propagation_distance, parameters, element_index=None):
        return self.do_specific_progation(wavefront, propagation_distance, parameters, prefix="after")

    def do_specific_progation(self, wavefront, propagation_distance, parameters, prefix="after"):
        is_generic_wavefront = isinstance(wavefront, GenericWavefront2D)

        if is_generic_wavefront:
            wavefront = SRWWavefront.fromGenericWavefront(wavefront)
        else:
            if not isinstance(wavefront, SRWWavefront): raise ValueError("wavefront cannot be managed by this propagator")

        #
        # propagation (simple wavefront drift
        #

        optBL = SRWLOptC([SRWLOptD(propagation_distance)], # drift space
                         [self.__get_drift_wavefront_propagation_parameters(parameters, prefix)])

        srwl.PropagElecField(wavefront, optBL)

        if is_generic_wavefront:
            return wavefront.toGenericWavefront()
        else:
            return wavefront

    def __get_drift_wavefront_propagation_parameters(self, parameters, where="before"):
        if not parameters.has_additional_parameter("srw_drift_" + where + "_wavefront_propagation_parameters"):
            wavefront_propagation_parameters = WavefrontPropagationParameters()
        else:
            wavefront_propagation_parameters = parameters.get_additional_parameter("srw_drift_" + where + "_wavefront_propagation_parameters")

            if not isinstance(wavefront_propagation_parameters, WavefrontPropagationParameters):
                raise ValueError("SRW Wavefront Propagation Parameters not present")

        srw_parameters_array = wavefront_propagation_parameters.to_SRW_array()

        if parameters.has_additional_parameter("srw_drift_" + where + "_wavefront_propagation_optional_parameters"):
            wavefront_propagation_optional_parameters = parameters.get_additional_parameter("srw_drift_" + where + "_wavefront_propagation_optional_parameters")

            if not isinstance(wavefront_propagation_optional_parameters, WavefrontPropagationOptionalParameters):
                raise ValueError("SRW Wavefront Propagation Optional Parameters are inconsistent")

            wavefront_propagation_optional_parameters.append_to_srw_array(srw_parameters_array)

        return srw_parameters_array



import numpy

from syned.beamline.shape import Ellipse, Rectangle
from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElement
import wofrysrw.util.srw_absorption as srwa

from wofrysrw.util.srw import *

class PlaneOfFocusing:
    HORIZONTAL=1
    VERTICAL=2
    BOTH=3

    @classmethod
    def items(cls):
        return ["Horizontal", "Vertical", "Both"]

class CRLShape:
    PARABOLIC=1
    SPHERICAL=2

    @classmethod
    def items(cls):
        return ["Parabolic", "Spherical"]

class SRWCRL(SRWOpticalElement):
    """
    Setup Transmission type Optical Element which simulates Compound Refractive Lens (CRL)
    :param _foc_plane: plane of focusing: 1- horizontal, 2- vertical, 3- both
    :param _delta: refractive index decrement (can be one number of array vs photon energy)
    :param _atten_len: attenuation length [m] (can be one number of array vs photon energy)
    :param _shape: 1- parabolic, 2- circular (spherical)
    :param _apert_h: horizontal aperture size [m]
    :param _apert_v: vertical aperture size [m]
    :param _r_min: radius (on tip of parabola for parabolic shape) [m]
    :param _n: number of lenses (/"holes")
    :param _wall_thick: min. wall thickness between "holes" [m]
    :param _xc: horizontal coordinate of center [m]
    :param _yc: vertical coordinate of center [m]
    :param _void_cen_rad: flat array/list of void center coordinates and radii: [x1, y1, r1, x2, y2, r2,...]
    :param _e_start: initial photon energy
    :param _e_fin: final photon energy
    :param _nx: number of points vs horizontal position to represent the transmission element
    :param _ny: number of points vs vertical position to represent the transmission element
    :param _ang_rot_ex: angle [rad] of CRL rotation about horizontal axis
    :param _ang_rot_ey: angle [rad] of CRL rotation about vertical axis
    :param _ang_rot_ez: angle [rad] of CRL rotation about longitudinal axis
    :return: transmission (SRWLOptT) type optical element which simulates CRL
    """
    def __init__(self,
                 name="Undefined",
                 optical_element_displacement       = None,
                 plane_of_focusing=PlaneOfFocusing.BOTH,
                 delta=1e-6,
                 attenuation_length=1e-3,
                 shape=CRLShape.PARABOLIC,
                 horizontal_aperture_size=1e-3,
                 vertical_aperture_size=1e-3,
                 radius_of_curvature=5e-3,
                 number_of_lenses=10,
                 wall_thickness=5e-5,
                 horizontal_center_coordinate=0.0,
                 vertical_center_coordinate=0.0,
                 void_center_coordinates=None,
                 initial_photon_energy=8000,
                 final_photon_energy=8010,
                 horizontal_points=1001,
                 vertical_points=1001,
                 thickness_error_profile_files=None,
                 scaling_factor=1.0):
        SRWOpticalElement.__init__(self, optical_element_displacement=optical_element_displacement)

        self._name = name

        self.plane_of_focusing = plane_of_focusing
        self.delta = delta
        self.attenuation_length = attenuation_length
        self.shape = shape
        self.horizontal_aperture_size = horizontal_aperture_size
        self.vertical_aperture_size = vertical_aperture_size
        self.radius_of_curvature = radius_of_curvature
        self.number_of_lenses = number_of_lenses
        self.wall_thickness = wall_thickness
        self.horizontal_center_coordinate = horizontal_center_coordinate
        self.vertical_center_coordinate = vertical_center_coordinate
        self.void_center_coordinates = void_center_coordinates
        self.initial_photon_energy = initial_photon_energy
        self.final_photon_energy = final_photon_energy
        self.horizontal_points = horizontal_points
        self.vertical_points = vertical_points

        if not thickness_error_profile_files is None:
            self.has_error = True
            thickness_profile = numpy.zeros((self.horizontal_points, self.vertical_points))

            x_range = [self.horizontal_center_coordinate - self.horizontal_aperture_size/2,
                       self.horizontal_center_coordinate + self.horizontal_aperture_size/2]
            y_range = [self.vertical_center_coordinate - self.vertical_aperture_size/2,
                       self.vertical_center_coordinate + self.vertical_aperture_size/2]

            for thickness_error_profile_file in thickness_error_profile_files:
                srwa.add_thickness_error_to_thickness_profile(thickness_profile,
                                                              thickness_error_profile_file,
                                                              scaling_factor,
                                                              x_range,
                                                              y_range,
                                                              horizontal_points,
                                                              vertical_points)
            self.error_transmission_amplitudes              = srwa.get_transmission_amplitudes(thickness_profile, attenuation_length)
            self.error_transmission_optical_path_difference = srwa.get_transmission_optical_path_difference(thickness_profile, delta)
        else:
            self.has_error = False

    def toSRWLOpt(self):
        crl_opt = srwl_opt_setup_CRL(_foc_plane=self.plane_of_focusing,
                                     _delta=self.delta,
                                     _atten_len=self.attenuation_length,
                                     _shape=self.shape,
                                     _apert_h=self.horizontal_aperture_size,
                                     _apert_v=self.vertical_aperture_size,
                                     _r_min=self.radius_of_curvature,
                                     _n=self.number_of_lenses,
                                     _wall_thick=self.wall_thickness,
                                     _xc=self.horizontal_center_coordinate,
                                     _yc=self.vertical_center_coordinate,
                                     _void_cen_rad=self.void_center_coordinates,
                                     _e_start=self.initial_photon_energy,
                                     _e_fin=self.final_photon_energy,
                                     _nx=self.horizontal_points,
                                     _ny=self.vertical_points)

        if self.has_error:
            srwa.add_thickness_error_transmission(crl_opt, self.horizontal_points, self.vertical_points,
                                                  self.error_transmission_amplitudes, self.error_transmission_optical_path_difference)

        return crl_opt

    def fromSRWLOpt(self, srwlopt=None):
        if not srwlopt.input_params or not srwlopt.input_params["type"] == "crl":
            raise TypeError("SRW optical element is not a CRL")

        self.plane_of_focusing = srwlopt.input_params["focalPlane"]
        self.delta = srwlopt.input_params["refractiveIndex"]
        self.attenuation_length = srwlopt.input_params["attenuationLength"]
        self.shape = srwlopt.input_params["shape"]
        self.horizontal_aperture_size = srwlopt.input_params["horizontalApertureSize"]
        self.vertical_aperture_size = srwlopt.input_params["verticalApertureSize"]
        self.radius_of_curvature = srwlopt.input_params["radius"]
        self.number_of_lenses = srwlopt.input_params["numberOfLenses"]
        self.wall_thickness = srwlopt.input_params["wallThickness"]
        self.horizontal_center_coordinate = srwlopt.input_params["horizontalCenterCoordinate"]
        self.vertical_center_coordinate = srwlopt.input_params["verticalCenterCoordinate"]
        self.void_center_coordinates = srwlopt.input_params["voidCenterCoordinates"]
        self.initial_photon_energy = srwlopt.input_params["initialPhotonEnergy"]
        self.final_photon_energy = srwlopt.input_params["finalPhotonEnergy"]
        self.horizontal_points = srwlopt.input_params["horizontalPoints"]
        self.vertical_points = srwlopt.input_params["verticalPoints"]

    def to_python_code(self, data=None):
        text_code = data[0] + "=srwl_opt_setup_CRL(_foc_plane=" + str(self.plane_of_focusing) + ",\n"
        text_code += "                _delta=" + str(self.delta) + ",\n"
        text_code += "                _atten_len=" + str(self.attenuation_length) + ",\n"
        text_code += "                _shape=" + str(self.shape) + ",\n"
        text_code += "                _apert_h=" + str(self.horizontal_aperture_size) + ",\n"
        text_code += "                _apert_v=" + str(self.vertical_aperture_size) + ",\n"
        text_code += "                _r_min=" + str(self.radius_of_curvature) + ",\n"
        text_code += "                _n=" + str(self.number_of_lenses) + ",\n"
        text_code += "                _wall_thick=" + str(self.wall_thickness) + ",\n"
        text_code += "                _xc=" + str(self.horizontal_center_coordinate) + ",\n"
        text_code += "                _yc=" + str(self.vertical_center_coordinate) + ",\n"
        text_code += "                _void_cen_rad=" + str(self.void_center_coordinates) + ",\n"
        text_code += "                _e_start=" + str(self.initial_photon_energy) + ",\n"
        text_code += "                _e_fin=" + str(self.final_photon_energy) + ",\n"
        text_code += "                _nx=" + str(self.horizontal_points) + ",\n"
        text_code += "                _ny=" + str(self.vertical_points) + ")\n"

        if self.has_error:
            text_code += "\n\n"
            text_code += "def add_thickness_error_transmission(srwlopt_t, n_points_x, n_points_y,\n" + \
                         "                                     error_transmission_amplitudes, error_transmission_optical_path_difference):\n" + \
                         "    ofst = 0\n" + \
                         "    for iy in range(n_points_y):\n" + \
                         "        for ix in range(n_points_x):\n" + \
                         "            srwlopt_t.arTr[ofst]     *= error_transmission_amplitudes[ix, iy]\n" + \
                         "            srwlopt_t.arTr[ofst + 1] += error_transmission_optical_path_difference[ix, iy]\n" + \
                         "            ofst += 2\n\n"

            def to_string(numpy_array):
                text = "["
                for i in range(numpy_array.shape[0]):
                    for j in range(numpy_array.shape[1]):
                        text += ("[" if j == 0 else "") + str(numpy_array[i, j]) + ("]" if j == numpy_array.shape[1]-1 else ", ")
                    text += ", "
                return text[:-2] + "]"

            print(to_string(self.error_transmission_optical_path_difference))

            text_code += "error_transmission_amplitudes              = numpy.array(" + to_string(self.error_transmission_amplitudes) + ")\n"
            text_code += "error_transmission_optical_path_difference = numpy.array(" + to_string(self.error_transmission_optical_path_difference) + ")\n\n"

            text_code += "add_thickness_error_transmission(" + \
                         data[0] + ", " + \
                         str(self.horizontal_points) + ", " + \
                         str(self.vertical_points) + ", error_transmission_amplitudes, error_transmission_optical_path_difference)\n\n"

        return text_code

    def get_boundary_shape(self):
        if self.plane_of_focusing == PlaneOfFocusing.BOTH:
            return Ellipse(a_axis_min=self.horizontal_center_coordinate - self.horizontal_aperture_size/2,
                           a_axis_max=self.horizontal_center_coordinate + self.horizontal_aperture_size/2,
                           b_axis_min=self.vertical_center_coordinate + self.vertical_aperture_size/2,
                           b_axis_max=self.vertical_center_coordinate + self.vertical_aperture_size/2)
        else:
            return Rectangle(x_left=self.horizontal_center_coordinate - self.horizontal_aperture_size/2,
                             x_right=self.horizontal_center_coordinate + self.horizontal_aperture_size/2,
                             y_bottom=self.vertical_center_coordinate + self.vertical_aperture_size/2,
                             y_top=self.vertical_center_coordinate + self.vertical_aperture_size/2)



import numpy

from syned.beamline.optical_elements.crystals.crystal import Crystal, DiffractionGeometry
from syned.beamline.shape import Ellipse, Rectangle, Circle, Plane

from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElement
from wofrysrw.propagator.wavefront2D.srw_wavefront import WavefrontPropagationParameters

from wofrysrw.util.srw import SRWLOptCryst
from wofrysrw.util.srw import srwl, srwl_opt_setup_surf_height_1d, srwl_opt_setup_surf_height_2d, srwl_uti_read_data_cols

from wofrysrw.beamline.optical_elements.mirrors.srw_mirror import Orientation, TreatInputOutput, ApertureShape, SimulationMethod

'''
        :param _d_sp: (_d_space) crystal reflecting planes d-spacing (John's dA) [A]
        :param _psi0r: real part of 0-th Fourier component of crystal polarizability (John's psi0c.real) (units?)
        :param _psi0i: imaginary part of 0-th Fourier component of crystal polarizability (John's psi0c.imag) (units?)
        :param _psi_hr: (_psiHr) real part of H-th Fourier component of crystal polarizability (John's psihc.real) (units?)
        :param _psi_hi: (_psiHi) imaginary part of H-th Fourier component of crystal polarizability (John's psihc.imag) (units?)
        :param _psi_hbr: (_psiHBr:) real part of -H-th Fourier component of crystal polarizability (John's psimhc.real) (units?)
        :param _psi_hbi: (_psiHBi:) imaginary part of -H-th Fourier component of crystal polarizability (John's psimhc.imag) (units?)
        :param _tc: crystal thickness [m] (John's thicum)
        :param _ang_as: (_Tasym) asymmetry angle [rad] (John's alphdg)
        :param _nvx: horizontal coordinate of outward normal to crystal surface (John's angles: thdg, chidg, phidg)
        :param _nvy: vertical coordinate of outward normal to crystal surface (John's angles: thdg, chidg, phidg)
        :param _nvz: longitudinal coordinate of outward normal to crystal surface (John's angles: thdg, chidg, phidg)
        :param _tvx: horizontal coordinate of central tangential vector (John's angles: thdg, chidg, phidg)
        :param _tvy: vertical coordinate of central tangential vector (John's angles: thdg, chidg, phidg)
        :param _uc: crystal use case: 1- Bragg Reflection, 2- Bragg Transmission (Laue cases to be added)

'''

class SRWCrystal(Crystal, SRWOpticalElement):
    def __init__(self,
                 name                                 = "Undefined",
                 optical_element_displacement         = None,
                 orientation_of_reflection_plane      = Orientation.UP,
                 invert_tangent_component             = False,
                 d_spacing                            = 0.0,
                 psi_0r                               = 0.0,
                 psi_0i                               = 0.0,
                 psi_hr                               = 0.0,
                 psi_hi                               = 0.0,
                 psi_hbr                              = 0.0,
                 psi_hbi                              = 0.0,
                 asymmetry_angle                      = 0.0,
                 thickness                            = 0.0,
                 diffraction_geometry                 = DiffractionGeometry.BRAGG,
                 incident_angle                       = 0.0
                ):
        SRWOpticalElement.__init__(self, optical_element_displacement=optical_element_displacement)

        Crystal.__init__(self,
                         name,
                         surface_shape=Plane(),
                         boundary_shape=Rectangle(x_left=0.0,
                                                  x_right=0.0,
                                                  y_bottom=0.0,
                                                  y_top=0.0),
                         material="Unknown",
                         diffraction_geometry=diffraction_geometry,
                         asymmetry_angle = asymmetry_angle,
                         thickness = thickness
                        )

        self.orientation_of_reflection_plane                  = orientation_of_reflection_plane
        self.invert_tangent_component                         = invert_tangent_component

        self.d_spacing                            = d_spacing
        self.psi_0r                               = psi_0r
        self.psi_0i                               = psi_0i
        self.psi_hr                               = psi_hr
        self.psi_hi                               = psi_hi
        self.psi_hbr                              = psi_hbr
        self.psi_hbi                              = psi_hbi
        self.asymmetry_angle                      = asymmetry_angle
        self.thickness                            = thickness
        self.diffraction_geometry                 = diffraction_geometry
        self.grazing_angle                        = incident_angle

        if diffraction_geometry == DiffractionGeometry.LAUE: raise NotImplementedError("Laue Geometry is not yet supported")

    def toSRWLOpt(self):
        nvx, nvy, nvz, tvx, tvy = self.get_orientation_vectors()

        return SRWLOptCryst(_d_sp=self.d_spacing,
                            _psi0r=self.psi_0r,
                            _psi0i=self.psi_0i,
                            _psi_hr=self.psi_hr,
                            _psi_hi=self.psi_hi,
                            _psi_hbr=self.psi_hbr,
                            _psi_hbi=self.psi_hbi,
                            _tc=self.thickness,
                            _ang_as=self.asymmetry_angle,
                            _nvx=nvx,
                            _nvy=nvy,
                            _nvz=nvz,
                            _tvx=tvx,
                            _tvy=tvy,
                            _uc=1 if self.diffraction_geometry==DiffractionGeometry.BRAGG else 0)

    def to_python_code(self, data=None):
        oe_name = data[0]

        nvx, nvy, nvz, tvx, tvy = self.get_orientation_vectors()

        text_code  = oe_name + "="+ "SRWLOptCryst(_d_sp=" + str(self.d_spacing) + "," + "\n"
        text_code += "                        _psi0r=" + str(self.psi_0r) + "," + "\n"
        text_code += "                        _psi0i=" + str(self.psi_0i) + "," + "\n"
        text_code += "                        _psi_hr=" + str(self.psi_hr) + "," + "\n"
        text_code += "                        _psi_hi=" + str(self.psi_hi) + "," + "\n"
        text_code += "                        _psi_hbr=" + str(self.psi_hbr) + "," + "\n"
        text_code += "                        _psi_hbi=" + str(self.psi_hbi) + "," + "\n"
        text_code += "                        _tc=" + str(self.thickness) + "," + "\n"
        text_code += "                        _ang_as=" + str(self.asymmetry_angle) + "," + "\n"
        text_code += "                        _nvx=" + str(nvx) + "," + "\n"
        text_code += "                        _nvy=" + str(nvy) + "," + "\n"
        text_code += "                        _nvz=" + str(nvz) + "," + "\n"
        text_code += "                        _tvx=" + str(tvx) + "," + "\n"
        text_code += "                        _tvy=" + str(tvy)+ "," + "\n"
        text_code += "                        _uc=" + str(1 if self.diffraction_geometry==DiffractionGeometry.BRAGG else 0) + ")" + "\n"

        return text_code



import numpy

from wofrysrw.util.srw import srwl


from wofrysrw.propagator.wavefront2D.srw_wavefront import WavefrontParameters
from wofrysrw.storage_ring.srw_light_source import SRWLightSource, PhotonSourceProperties
from wofrysrw.storage_ring.srw_electron_beam import SRWElectronBeam
from wofrysrw.storage_ring.magnetic_structures.srw_undulator import SRWUndulator

import scipy.constants as codata
m2ev = codata.c * codata.h / codata.e      # lambda(m)  = m2eV / energy(eV)

class FluxPrecisionParameters(object):
    def __init__(self,
                 initial_UR_harmonic = 1,
                 final_UR_harmonic = 21, #Maximum number of harmonics considered. This is critical for speed
                 longitudinal_integration_precision_parameter = 1.5,
                 azimuthal_integration_precision_parameter = 1.5,
                 calculation_type = 2 #calculate flux (1) or flux per unit surface (2)
                 ):
        self._initial_UR_harmonic = initial_UR_harmonic
        self._final_UR_harmonic = final_UR_harmonic
        self._longitudinal_integration_precision_parameter = longitudinal_integration_precision_parameter
        self._azimuthal_integration_precision_parameter = azimuthal_integration_precision_parameter
        self._calculation_type = calculation_type

    def to_SRW_array(self):
        return [int(self._initial_UR_harmonic),
                int(self._final_UR_harmonic),
                float(self._longitudinal_integration_precision_parameter),
                float(self._azimuthal_integration_precision_parameter),
                int(self._calculation_type)]

class SRWUndulatorLightSource(SRWLightSource):

    def __init__(self,
                 name="Undefined",
                 electron_beam=SRWElectronBeam(),
                 undulator_magnetic_structure=SRWUndulator()):

        super().__init__(name,
                         electron_beam=electron_beam,
                         magnetic_structure=undulator_magnetic_structure)

    def get_length(self):
        return self._magnetic_structure._period_length*self._magnetic_structure._number_of_periods

    def get_resonance_wavelength(self):
        return self._magnetic_structure.resonance_wavelength(self._electron_beam.gamma())

    def get_resonance_energy(self):
        return m2ev / self.get_resonance_wavelength()

    def get_photon_source_properties(self, harmonic):
        wavelength = m2ev/(harmonic*self.get_resonance_energy())
        undulator_length = self.get_length()

        # calculate sizes of the photon undulator beam
        # see formulas 25 & 30 in Elleaume (Onaki & Elleaume)
        s_phot = 2.740/(4e0*numpy.pi)*numpy.sqrt(undulator_length*wavelength)
        sp_phot = 0.69*numpy.sqrt(wavelength/undulator_length)

        photon_h = numpy.sqrt(numpy.power(numpy.sqrt(self._electron_beam._moment_xx), 2) + numpy.power(s_phot, 2))
        photon_v = numpy.sqrt(numpy.power(numpy.sqrt(self._electron_beam._moment_yy), 2) + numpy.power(s_phot, 2))
        photon_hp = numpy.sqrt(numpy.power(numpy.sqrt(self._electron_beam._moment_xpxp), 2) + numpy.power(sp_phot, 2))
        photon_vp = numpy.sqrt(numpy.power(numpy.sqrt(self._electron_beam._moment_ypyp), 2) + numpy.power(sp_phot, 2))

        cohH = wavelength/4/numpy.pi / photon_h / photon_hp
        cohV = wavelength/4/numpy.pi / photon_v / photon_vp

        dls = numpy.sqrt(2*undulator_length*wavelength)/4/numpy.pi

        return PhotonSourceProperties(rms_h=photon_h,
                                      rms_v=photon_v,
                                      rms_hp=photon_hp,
                                      rms_vp=photon_vp,
                                      coherence_volume_h=cohH,
                                      coherence_volume_v=cohV,
                                      diffraction_limit=dls)

    def get_undulator_flux(self,
                           source_wavefront_parameters = WavefrontParameters(),
                           flux_precision_parameters = FluxPrecisionParameters()):

        stkF = source_wavefront_parameters.to_SRWLStokes()

        srwl.CalcStokesUR(stkF,
                          self._electron_beam.to_SRWLPartBeam(),
                          self._magnetic_structure.get_SRWMagneticStructure(),
                          flux_precision_parameters.to_SRW_array())

        eArray = numpy.zeros(stkF.mesh.ne)
        intensArray = numpy.zeros(stkF.mesh.ne)

        for i in range(source_wavefront_parameters._photon_energy_points):
            eArray[i] = stkF.mesh.eStart+i*(stkF.mesh.eFin-stkF.mesh.eStart)/numpy.array((stkF.mesh.ne-1)).clip(min=1)
            intensArray[i] = stkF.arS[i]

        return (eArray, intensArray)




import scipy.constants as codata
angstroms_to_eV = codata.h*codata.c/codata.e*1e10

from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
from wofry.propagator.propagator import PropagationManager, PropagationParameters, Propagator2D

from wofrysrw.beamline.srw_beamline import Where
from wofrysrw.propagator.wavefront2D.srw_wavefront import WavefrontPropagationParameters, WavefrontPropagationOptionalParameters
from wofrysrw.propagator.wavefront2D.srw_wavefront import SRWWavefront
from wofrysrw.propagator.propagators2D.srw_propagation_mode import SRWPropagationMode

from wofrysrw.util.srw import *

SRW_APPLICATION = "SRW"

class FresnelSRWNative(Propagator2D):

    HANDLER_NAME = "FRESNEL_SRW_NATIVE"

    def get_handler_name(self):
        return self.HANDLER_NAME

    """
    2D Fresnel propagator using convolution via Fourier transform
    :param wavefront:
    :param propagation_distance:
    :param srw_autosetting:set to 1 for automatic SRW redimensionate wavefront
    :return:
    """

    def do_propagation(self, parameters=PropagationParameters()):
        wavefront = parameters.get_wavefront()

        is_generic_wavefront = isinstance(wavefront, GenericWavefront2D)

        if is_generic_wavefront:
            wavefront = SRWWavefront.fromGenericWavefront(wavefront)
        else:
            if not isinstance(wavefront, SRWWavefront): raise ValueError("wavefront cannot be managed by this propagator")

        srw_oe_array = []
        srw_pp_array = []

        propagation_mode = PropagationManager.Instance().get_propagation_mode(SRW_APPLICATION)

        if propagation_mode == SRWPropagationMode.STEP_BY_STEP:
            self.add_optical_element(parameters, 0, srw_oe_array, srw_pp_array, wavefront)
        elif propagation_mode == SRWPropagationMode.WHOLE_BEAMLINE:
            srw_beamline = parameters.get_additional_parameter("working_beamline")

            for index in range(srw_beamline.get_beamline_elements_number()):
                self.add_optical_element_from_beamline(srw_beamline, index, srw_oe_array, srw_pp_array, wavefront)
        else:
            raise ValueError("Propagation Mode not supported by this Propagator")

        if len(srw_oe_array) > 0:
            optBL = SRWLOptC(srw_oe_array, srw_pp_array)
            srwl.PropagElecField(wavefront, optBL)

        if is_generic_wavefront:
            return wavefront.toGenericWavefront()
        else:
            return wavefront

    ########################################################
    # WHOLE BEAMLINE

    def add_optical_element_from_beamline(self, srw_beamline, index, srw_oe_array, srw_pp_array, wavefront):
        optical_element = srw_beamline.get_beamline_element_at(index).get_optical_element()
        coordinates = srw_beamline.get_beamline_element_at(index).get_coordinates()

        if coordinates.p() != 0.0:
            srw_oe_array.append(SRWLOptD(coordinates.p()))
            srw_pp_array.append(self.__get_drift_wavefront_propagation_parameters_from_beamline(srw_beamline, index, Where.DRIFT_BEFORE))

        optical_element.add_to_srw_native_array(srw_oe_array, srw_pp_array, srw_beamline.get_wavefront_propagation_parameters_at(index, Where.OE), wavefront)

        if coordinates.q() != 0.0:
            srw_oe_array.append(SRWLOptD(coordinates.q()))
            srw_pp_array.append(self.__get_drift_wavefront_propagation_parameters_from_beamline(srw_beamline, index, Where.DRIFT_AFTER))

    ########################################################
    # ELEMENT BY ELEMENT

    def __get_drift_wavefront_propagation_parameters_from_beamline(self, srw_beamline, index, where=Where.DRIFT_BEFORE):
        wavefront_propagation_parameters, wavefront_propagation_optional_parameters = srw_beamline.get_wavefront_propagation_parameters_at(index, where)

        srw_parameters_array = wavefront_propagation_parameters.to_SRW_array()
        if not wavefront_propagation_optional_parameters is None: wavefront_propagation_optional_parameters.append_to_srw_array(srw_parameters_array)

        return srw_parameters_array

    def __get_drift_wavefront_propagation_parameters(self, parameters, where=Where.DRIFT_BEFORE):
        if not parameters.has_additional_parameter("srw_drift_" + where + "_wavefront_propagation_parameters"):
            wavefront_propagation_parameters = WavefrontPropagationParameters()
        else:
            wavefront_propagation_parameters = parameters.get_additional_parameter("srw_drift_" + where + "_wavefront_propagation_parameters")

            if not isinstance(wavefront_propagation_parameters, WavefrontPropagationParameters):
                raise ValueError("SRW Wavefront Propagation Parameters not present")

        srw_parameters_array = wavefront_propagation_parameters.to_SRW_array()

        if parameters.has_additional_parameter("srw_drift_" + where + "_wavefront_propagation_optional_parameters"):
            wavefront_propagation_optional_parameters = parameters.get_additional_parameter("srw_drift_" + where + "_wavefront_propagation_optional_parameters")

            if not isinstance(wavefront_propagation_optional_parameters, WavefrontPropagationOptionalParameters):
                raise ValueError("SRW Wavefront Propagation Optional Parameters are inconsistent")

            wavefront_propagation_optional_parameters.append_to_srw_array(srw_parameters_array)

        return srw_parameters_array

    def add_optical_element(self, parameters, index, srw_oe_array, srw_pp_array, wavefront):
        optical_element = parameters.get_PropagationElements().get_propagation_element(index).get_optical_element()
        coordinates = parameters.get_PropagationElements().get_propagation_element(index).get_coordinates()

        if coordinates.p() != 0.0:
            srw_oe_array.append(SRWLOptD(coordinates.p()))
            srw_pp_array.append(self.__get_drift_wavefront_propagation_parameters(parameters, Where.DRIFT_BEFORE))

        optical_element.add_to_srw_native_array(srw_oe_array, srw_pp_array, parameters, wavefront)

        if coordinates.q() != 0.0:
            srw_oe_array.append(SRWLOptD(coordinates.q()))
            srw_pp_array.append(self.__get_drift_wavefront_propagation_parameters(parameters, Where.DRIFT_AFTER))



import numpy


from syned.beamline.optical_elements.absorbers.filter import Filter
from wofrysrw.beamline.optical_elements.absorbers.srw_transmission import SRWTransmission
from wofrysrw.util.srw_absorption import get_transmission_amplitudes, get_transmission_optical_path_difference, add_thickness_error_to_thickness_profile

class SRWFilter(Filter, SRWTransmission):
    def __init__(self,
                 name="Undefined",
                 material="Be",
                 thickness=1e-3,
                 attenuation_length=1.0,
                 delta=1e-6,
                 x_range=[-1e-3, 1e-3],
                 y_range=[-1e-3, 1e-3],
                 n_points_x=100,
                 n_points_y=100,
                 energy=15000,
                 thickness_error_profile_file=None,
                 scaling_factor=1.0):
        Filter.__init__(self, name=name, material=material, thickness=thickness)

        thickness_profile = numpy.ones((n_points_x, n_points_y))*self.get_thickness()

        if not thickness_error_profile_file is None: add_thickness_error_to_thickness_profile(thickness_profile,
                                                                                              thickness_error_profile_file,
                                                                                              scaling_factor,
                                                                                              x_range,
                                                                                              y_range,
                                                                                              n_points_x,
                                                                                              n_points_y)

        SRWTransmission.__init__(self,
                                 x_range=x_range,
                                 y_range=y_range,
                                 transmission_amplitudes=get_transmission_amplitudes(thickness_profile, attenuation_length),
                                 transmission_optical_path_difference=get_transmission_optical_path_difference(thickness_profile, delta),
                                 energy=energy)







from syned.beamline.beamline import Beamline

from wofrysrw.srw_object import SRWObject
from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElementDisplacement
from wofrysrw.storage_ring.srw_light_source import SRWLightSource
from wofrysrw.propagator.wavefront2D.srw_wavefront import WavefrontPropagationParameters, WavefrontPropagationOptionalParameters

class Where:
    DRIFT_BEFORE = "before"
    DRIFT_AFTER = "after"
    OE = "oe"

    @classmethod
    def tuple(cls):
        return cls.DRIFT_BEFORE, cls.OE, cls.DRIFT_AFTER

class SRWBeamline(Beamline, SRWObject):

    def __init__(self,
                 light_source=SRWLightSource(),
                 beamline_elements_list=[]):
        super().__init__(light_source=light_source, beamline_elements_list=beamline_elements_list)

        self._wavefront_propagation_parameters_list = {}
        self._wavefront_propagation_parameters_list[Where.DRIFT_BEFORE] = []
        self._wavefront_propagation_parameters_list[Where.OE]           = []
        self._wavefront_propagation_parameters_list[Where.DRIFT_AFTER]  = []

    # overwrites the SynedObject method for dealing with list
    def to_dictionary(self): # TODO: to be adjusted....
        dict_to_save = super().to_dictionary()
        dict_to_save["srw_wavefront_propagation_parameters"] = [ el[0].to_dictionary() for el in self._wavefront_propagation_parameters_list[Where.OE]]

        return dict_to_save

    def append_wavefront_propagation_parameters(self, wavefront_propagation_parameters=WavefrontPropagationParameters(), wavefront_propagation_optional_parameters=WavefrontPropagationOptionalParameters(), where=Where.OE):
        self._wavefront_propagation_parameters_list[where].append([wavefront_propagation_parameters, wavefront_propagation_optional_parameters])

    def get_wavefront_propagation_parameters(self, where=Where.OE):
        return self._wavefront_propagation_parameters_list[where]

    def get_wavefront_propagation_parameters_at(self, index, where=Where.OE):
        if index >= len(self._wavefront_propagation_parameters_list[where]):
            raise IndexError("Index " + str(index) + " out of bounds")

        return self._wavefront_propagation_parameters_list[where][index]

    def duplicate(self):
        beamline_elements_list = []
        for beamline_element in self._beamline_elements_list:
            beamline_elements_list.append(beamline_element)

        new_beamline = SRWBeamline(light_source=self._light_source, beamline_elements_list = beamline_elements_list)

        for where in Where.tuple():
            for element in self.get_wavefront_propagation_parameters(where):
                new_beamline.append_wavefront_propagation_parameters(element[0], element[1], where)

        return new_beamline

    def to_python_code(self, data=None):
        wavefront = data[0]
        is_multi_electron = data[1] == True


        text_code =  "try:\n"
        text_code += "    from srwpy.srwlib import *\n    from srwpy.uti_plot import *\n"
        text_code  += "except: \n"
        text_code += "    from wofrysrw.util.srw import *\n"
        text_code += "import numpy\n\n"

        if is_multi_electron: text_code += "#if not srwl_uti_proc_is_master(): exit()\n"
        else: text_code += "if not srwl_uti_proc_is_master(): exit()\n"

        text_code += "\n####################################################\n# LIGHT SOURCE\n\n"
        text_code += self.get_light_source().to_python_code(is_multi_electron)

        if not is_multi_electron:
            text_code += "\n"
            text_code += "mesh0 = deepcopy(wfr.mesh)" + "\n"
            text_code += "arI = array('f', [0]*mesh0.nx*mesh0.ny)" + "\n"
            text_code += "srwl.CalcIntFromElecField(arI, wfr, 6, 0, 3, mesh0.eStart, 0, 0)" + "\n"
            text_code += "arIx = array('f', [0]*mesh0.nx)" + "\n"
            text_code += "srwl.CalcIntFromElecField(arIx, wfr, 6, 0, 1, mesh0.eStart, 0, 0)" + "\n"
            text_code += "arIy = array('f', [0]*mesh0.ny)" + "\n"
            text_code += "srwl.CalcIntFromElecField(arIy, wfr, 6, 0, 2, mesh0.eStart, 0, 0)" + "\n"
            text_code += "#save ascii file with intensity" + "\n"
            text_code += "#srwl_uti_save_intens_ascii(arI, mesh0, <file_path>)" + "\n"
            text_code += "plotMesh0x = [1000*mesh0.xStart, 1000*mesh0.xFin, mesh0.nx]" + "\n"
            text_code += "plotMesh0y = [1000*mesh0.yStart, 1000*mesh0.yFin, mesh0.ny]" + "\n"
            text_code += "uti_plot_init(backend=\"Qt5Agg\")" + "\n"
            text_code += "uti_plot2d1d (arI, plotMesh0x, plotMesh0y, labels=['Horizontal Position [mm]', 'Vertical Position [mm]', 'Intensity Before Propagation'])" + "\n"

        if self.get_beamline_elements_number() > 0:
            text_code += "\n####################################################\n# BEAMLINE\n\n"
            text_code += "srw_oe_array = []" + "\n"
            text_code += "srw_pp_array = []" + "\n"
            text_code += "\n"

            for index in range(self.get_beamline_elements_number()):
                oe_name = "oe_" + str(index)
                data_oe = [oe_name, wavefront]

                beamline_element = self.get_beamline_element_at(index)

                optical_element = beamline_element.get_optical_element()
                coordinates = beamline_element.get_coordinates()

                # DRIFT BEFORE ----------------

                if coordinates.p() != 0.0:
                    text_code += "drift_before_" + oe_name + " = " + "SRWLOptD(" + str(coordinates.p()) + ")" + "\n"

                    wp, wop = self.get_wavefront_propagation_parameters_at(index, Where.DRIFT_BEFORE)
                    drift_pp_array = wp.to_python_code()
                    if not wop is None: drift_pp_array = wop.append_to_python_code(drift_pp_array)

                    text_code += "pp_drift_before_" + oe_name + " = " + drift_pp_array  + "\n"
                    text_code += "\n"
                    text_code += "srw_oe_array.append(drift_before_" + oe_name + ")" + "\n"
                    text_code += "srw_pp_array.append(pp_drift_before_" + oe_name + ")" + "\n"
                    text_code += "\n"

                wp, wop = self.get_wavefront_propagation_parameters_at(index, Where.OE)

                # OPTICAL ELEMENT ----------------

                if not wp is None: # SCREEN!!!
                    text_code += optical_element.to_python_code(data_oe)

                    oe_pp_array = wp.to_python_code()
                    if not wop is None: oe_pp_array = wop.append_to_python_code(oe_pp_array)

                    text_code += "\n"

                    if not optical_element.displacement is None:
                        data_oe.append(SRWOpticalElementDisplacement.BEFORE)

                        text_code += optical_element.displacement.to_python_code(data_oe)

                        text_code += "\n"

                        if optical_element.displacement.has_shift:
                            text_code += "srw_oe_array.append(shift_before_" + oe_name + ")" + "\n"
                            text_code += "srw_pp_array.append(pp_shift_before_" + oe_name + ")" + "\n"
                            text_code += "\n"

                        if optical_element.displacement.has_rotation:
                            text_code += "srw_oe_array.append(rotation_before_" + oe_name + ")" + "\n"
                            text_code += "srw_pp_array.append(pp_rotation_before_" + oe_name + ")" + "\n"
                            text_code += "\n"

                    if hasattr(optical_element, "add_acceptance_slit") and getattr(optical_element, "add_acceptance_slit") == True: # MIRROR AND GRATINGS

                        text_code += "pp_acceptance_slits_" + oe_name + " = " + oe_pp_array  + "\n"
                        text_code += "pp_" + oe_name + " = " + WavefrontPropagationParameters().to_python_code()  + "\n"
                        text_code += "\n"

                        text_code += "srw_oe_array.append(acceptance_slits_" + oe_name + ")" + "\n"
                        text_code += "srw_pp_array.append(pp_acceptance_slits_" + oe_name + ")" + "\n"

                    else:
                        text_code += "pp_" + oe_name + " = " + oe_pp_array  + "\n"

                    text_code += "\n"
                    text_code += "srw_oe_array.append(" + oe_name + ")" + "\n"
                    text_code += "srw_pp_array.append(pp_" + oe_name + ")" + "\n"

                    if hasattr(optical_element, "height_profile_data_file"): # MIRROR AND GRATINGS
                        if not getattr(optical_element, "height_profile_data_file") is None:
                            text_code += "\n"
                            text_code += "srw_oe_array.append(optTrEr_" + oe_name + ")" + "\n"
                            text_code += "srw_pp_array.append(" + WavefrontPropagationParameters().to_python_code() + ")" + "\n"

                    if not optical_element.displacement is None:
                        data_oe[2] = SRWOpticalElementDisplacement.AFTER

                        text_code += "\n"

                        text_code += optical_element.displacement.to_python_code(data_oe)

                        text_code += "\n"

                        if optical_element.displacement.has_shift:
                            text_code += "srw_oe_array.append(shift_after_" + oe_name + ")" + "\n"
                            text_code += "srw_pp_array.append(pp_shift_after_" + oe_name + ")" + "\n"
                            text_code += "\n"

                        if optical_element.displacement.has_rotation:
                            text_code += "srw_oe_array.append(rotation_after_" + oe_name + ")" + "\n"
                            text_code += "srw_pp_array.append(pp_rotation_after_" + oe_name + ")" + "\n"
                            text_code += "\n"


                # DRIFT AFTER ----------------

                if coordinates.q() != 0.0:
                    text_code += "\n"
                    text_code += "drift_after_" + oe_name + " = " + "SRWLOptD(" + str(coordinates.q()) + ")" + "\n"

                    wp, wop = self.get_wavefront_propagation_parameters_at(index, Where.DRIFT_AFTER)
                    drift_pp_array = wp.to_python_code()
                    if not wop is None: drift_pp_array = wop.append_to_python_code(drift_pp_array)

                    text_code += "pp_drift_after_" + oe_name + " = " + drift_pp_array  + "\n"
                    text_code += "\n"
                    text_code += "srw_oe_array.append(drift_after_" + oe_name + ")" + "\n"
                    text_code += "srw_pp_array.append(pp_drift_after_" + oe_name + ")" + "\n"

                text_code += "\n"

            text_code += "\n####################################################\n# PROPAGATION\n\n"
            text_code += "optBL = SRWLOptC(srw_oe_array, srw_pp_array)" + "\n"
            if not is_multi_electron: text_code += "srwl.PropagElecField(wfr, optBL)" + "\n"
            text_code += "\n"

            if not is_multi_electron:
                text_code += "mesh1 = deepcopy(wfr.mesh)" + "\n"
                text_code += "arI1 = array('f', [0]*mesh1.nx*mesh1.ny)" + "\n"
                text_code += "srwl.CalcIntFromElecField(arI1, wfr, 6, 0, 3, mesh1.eStart, 0, 0)" + "\n"
                text_code += "arI1x = array('f', [0]*mesh1.nx)" + "\n"
                text_code += "srwl.CalcIntFromElecField(arI1x, wfr, 6, 0, 1, mesh1.eStart, 0, 0)" + "\n"
                text_code += "arI1y = array('f', [0]*mesh1.ny)" + "\n"
                text_code += "srwl.CalcIntFromElecField(arI1y, wfr, 6, 0, 2, mesh1.eStart, 0, 0)" + "\n"
                text_code += "#save ascii file with intensity" + "\n"
                text_code += "#srwl_uti_save_intens_ascii(arI1, mesh1, <file_path>)" + "\n"
                text_code += "plotMesh1x = [1000*mesh1.xStart, 1000*mesh1.xFin, mesh1.nx]" + "\n"
                text_code += "plotMesh1y = [1000*mesh1.yStart, 1000*mesh1.yFin, mesh1.ny]" + "\n"
                text_code += "uti_plot2d1d(arI1, plotMesh1x, plotMesh1y, labels=['Horizontal Position [mm]', 'Vertical Position [mm]', 'Intensity After Propagation'])" + "\n"

        if not is_multi_electron: text_code += "uti_plot_show()" + "\n"

        if is_multi_electron:
            parameters = data[2]

            sampFactNxNyForProp  = parameters[0]
            nMacroElec           = parameters[1]
            nMacroElecAvgOneProc = parameters[2]
            nMacroElecSavePer    = parameters[3]
            srCalcMeth           = parameters[4]
            srCalcPrec           = parameters[5]
            strIntPropME_OutFileName = parameters[6]
            _char = parameters[7]

            text_code += "\n\n####################################################\n# MULTI ELECTRON PROPAGATION\n\n"

            text_code += "radStokesProp = srwl_wfr_emit_prop_multi_e(part_beam," + "\n"
            text_code += "                                           magnetic_field_container," + "\n"
            text_code += "                                           initial_mesh," + "\n"
            text_code += "                                           " + str(srCalcMeth) + "," + "\n"
            text_code += "                                           " + str(srCalcPrec) + "," + "\n"
            text_code += "                                           " + str(nMacroElec) + "," + "\n"
            text_code += "                                           " + str(nMacroElecAvgOneProc) + "," + "\n"
            text_code += "                                           " + str(nMacroElecSavePer) + "," + "\n"
            text_code += "                                           '" + str(strIntPropME_OutFileName) + "'," + "\n"
            text_code += "                                           " + str(sampFactNxNyForProp) + "," + "\n"
            text_code += "                                           optBL," + "\n"
            text_code += "                                           _char=" + str(int(_char)) + ")" + "\n"

        return text_code




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

from __future__ import print_function #Python 2.7 compatibility

__authors__ = ["R Celestre, M Sanchez del Rio, M Glass - ESRF"]
__license__ = "MIT"
__date__ = "27/05/2018"




from wofrysrw.util.srw import *
import numpy
import h5py

import time
import sys
import os


def save_wfr_2_hdf5(wfr,filename,subgroupname="wfr",intensity=False,phase=False,overwrite=True):
    """
    Writes wavefront data into a hdf5 generic file.
    When using the append mode to write h5 files, overwriting forces to initializes a new file.
    :param wfr: input / output resulting Wavefront structure (instance of SRWLWfr);
    :param filename: path to file for saving the wavefront
    :param subgroupname: container mechanism by which HDF5 files are organised
    :param intensity: Single-Electron" Intensity - Possible values:
            0 or False: Do not write intensity (Default)
            1 or True: Writes total intensity (total polarisation)
            2: Writes total intensity (total polarisation) plus sigma polarization and pi polarization
    :param phase: "Single-Electron" Radiation Phase - total polarisation (instance of srwl.CalcIntFromElecField)
    :param overwrite: flag that should always be set to True to avoid infinity loop on the recursive part of the function.
    """

    _complex_amplitude=True

    if (os.path.isfile(filename)) and (overwrite==True):
        os.remove(filename)
        FileName = filename.split("/")
        print("save_wfr_2_hdf5: file deleted %s"%FileName[-1])

    if not os.path.isfile(filename):  # if file doesn't exist, create it.
        sys.stdout.flush()
        f = h5py.File(filename, 'w')
        # points to the default data to be plotted
        f.attrs['default']          = 'entry'
        # give the HDF5 root some more attributes
        f.attrs['file_name']        = filename
        f.attrs['file_time']        = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
        f.attrs['creator']          = 'save_wfr_2_hdf5'
        f.attrs['code']             = 'SRW'
        f.attrs['HDF5_Version']     = h5py.version.hdf5_version
        f.attrs['h5py_version']     = h5py.version.version
        f.close()


    if phase:
        # s
        ar1 = array('d', [0] * wfr.mesh.nx * wfr.mesh.ny)  # "flat" 2D array to take intensity data
        srwl.CalcIntFromElecField(ar1, wfr, 0, 4, 3, wfr.mesh.eStart, 0, 0)
        arxx = numpy.array(ar1)
        arxx = arxx.reshape((wfr.mesh.ny, wfr.mesh.nx))#.T


        # p
        ar2 = array('d', [0] * wfr.mesh.nx * wfr.mesh.ny)  # "flat" 2D array to take intensity data
        srwl.CalcIntFromElecField(ar2, wfr, 1, 4, 3, wfr.mesh.eStart, 0, 0)
        aryy = numpy.array(ar2)
        aryy = aryy.reshape((wfr.mesh.ny, wfr.mesh.nx))#.T

        _dump_arr_2_hdf5(arxx-aryy, "phase/wfr_phase", filename, subgroupname) # difference
        _dump_arr_2_hdf5(arxx,      "phase/wfr_phase_s", filename, subgroupname)
        _dump_arr_2_hdf5(aryy,      "phase/wfr_phase_p", filename, subgroupname)

    if (_complex_amplitude) or (intensity):
        x_polarization = _SRWArrayToNumpy(wfr.arEx, wfr.mesh.nx, wfr.mesh.ny, wfr.mesh.ne)   # sigma
        y_polarization = _SRWArrayToNumpy(wfr.arEy, wfr.mesh.nx, wfr.mesh.ny, wfr.mesh.ne)   # pi

        e_field = numpy.concatenate((x_polarization, y_polarization), 3)

        complex_amplitude_s = e_field[0,:,:,0]
        complex_amplitude_p = e_field[0,:,:,1]

        if _complex_amplitude:
            _dump_arr_2_hdf5(complex_amplitude_s.T, "wfr_complex_amplitude_s", filename, subgroupname)
            _dump_arr_2_hdf5(complex_amplitude_p.T, "wfr_complex_amplitude_p", filename, subgroupname)

        if intensity:
            intens_p = numpy.abs(complex_amplitude_p) ** 2
            intens_s = numpy.abs(complex_amplitude_s) ** 2
            intens = intens_s + intens_p

            _dump_arr_2_hdf5(intens.T, "intensity/wfr_intensity", filename, subgroupname)
            if intensity == 2:
                _dump_arr_2_hdf5(intens_s.T, "intensity/wfr_intensity_s", filename, subgroupname)
                _dump_arr_2_hdf5(intens_p.T, "intensity/wfr_intensity_p", filename, subgroupname)

    f = h5py.File(filename, 'a')
    f1 = f[subgroupname]

    f2 = f1.create_group("partBeam")
    f2.attrs['NX_class'] = 'NXdata'
    f2["Iavg"]           = wfr.partBeam.Iavg
    f2["nPart"]          = wfr.partBeam.nPart
    f2["particle_x"]     = wfr.partBeam.partStatMom1.x
    f2["particle_y"]     = wfr.partBeam.partStatMom1.y
    f2["particle_z"]     = wfr.partBeam.partStatMom1.z
    f2["particle_xp"]    = wfr.partBeam.partStatMom1.xp
    f2["particle_yp"]    = wfr.partBeam.partStatMom1.yp
    f2["particle_gamma"] = wfr.partBeam.partStatMom1.gamma
    f2["particle_relE0"] = wfr.partBeam.partStatMom1.relE0
    f2["particle_nq"]    = wfr.partBeam.partStatMom1.nq
    f2.create_dataset('arStatMom2', data=numpy.array(wfr.partBeam.arStatMom2))

    # points to the default data to be plotted
    f1.attrs['NX_class'] = 'NXentry'
    f1.attrs['default']  = 'intensity'

    f1["wfr_typeE"]         = wfr.numTypeElFld
    f1["wfr_xc"]            = wfr.xc
    f1["wfr_yc"]            = wfr.yc
    f1["wfr_avgPhotEn"]     = wfr.avgPhotEn
    f1["wfr_presCA"]        = wfr.presCA
    f1["wfr_presFT"]        = wfr.presFT
    f1["wfr_unitElFld"]     = wfr.unitElFld
    f1["wfr_unitElFldAng"]  = wfr.unitElFldAng

    f1["wfr_Rx_dRx"] = numpy.array([wfr.Rx, wfr.dRx])
    f1["wfr_Ry_dRy"] = numpy.array([wfr.Ry, wfr.dRy])

    f1["wfr_photon_energy"] = wfr.mesh.eStart
    f1["wfr_zStart"]        = wfr.mesh.zStart
    f1["wfr_mesh_X"]        = numpy.array([wfr.mesh.xStart, wfr.mesh.xFin, wfr.mesh.nx])
    f1["wfr_mesh_Y"]        = numpy.array([wfr.mesh.yStart, wfr.mesh.yFin, wfr.mesh.ny])
    f1["wfr_mesh_E"]        = numpy.array([wfr.mesh.eStart, wfr.mesh.eFin, wfr.mesh.ne])

    _dump_arr_2_hdf5(numpy.array(wfr.arElecPropMatr), "wfr_arElecPropMatr", filename, subgroupname)
    _dump_arr_2_hdf5(numpy.array(wfr.arMomX),         "wfr_arMomX",         filename, subgroupname)
    _dump_arr_2_hdf5(numpy.array(wfr.arMomY),         "wfr_arMomY",         filename, subgroupname)
    _dump_arr_2_hdf5(numpy.array(wfr.arWfrAuxData),   "wfr_arWfrAuxData",   filename, subgroupname)


    # Add NX plot attribites for automatic plot with silx view
    myflags = [intensity, phase]
    mylabels = ['intensity', 'phase']
    for i,label in enumerate(mylabels):
        if myflags[i]:
            f2 = f1[mylabels[i]]
            f2.attrs['NX_class'] = 'NXdata'
            f2.attrs['signal'] = 'wfr_%s'%(mylabels[i])
            f2.attrs['axes'] = [b'axis_y', b'axis_x']

            # ds = nxdata.create_dataset('image_data', data=data)
            f3 = f2["wfr_%s"%(mylabels[i])]
            f3.attrs['interpretation'] = 'image'

            # X axis data
            ds = f2.create_dataset('axis_y', data=1e6*numpy.linspace(wfr.mesh.yStart,wfr.mesh.yFin,wfr.mesh.ny))
            # f1['axis1_name'] = numpy.arange(_wfr.mesh.ny)
            ds.attrs['units'] = 'microns'
            ds.attrs['long_name'] = 'Y Pixel Size (microns)'    # suggested X axis plot label
            #
            # Y axis data
            ds = f2.create_dataset('axis_x', data=1e6*numpy.linspace(wfr.mesh.xStart,wfr.mesh.xFin,wfr.mesh.nx))
            ds.attrs['units'] = 'microns'
            ds.attrs['long_name'] = 'X Pixel Size (microns)'    # suggested Y axis plot label
    f.close()

    FileName = filename.split("/")
    print("save_wfr_2_hdf5: file written/updated %s" %FileName[-1])

def load_hdf5_2_wfr(filename,filepath):
    """
    Loads a wawefront from an hdf5 file into a SRW wavefront object.

    :param filename: the file name where a SRW wavefront has been dumped
    :param filepath: the trying to access the file (wavefront entry name in the file, e.g., "wfr" or "wfr_end")
    :return:
    """
    wdic = load_hdf5_2_dictionary(filename, filepath)
    wfr = _dictionary_to_wfr(wdic)
    return wfr

def save_stokes_2_hdf5(_Stokes,_filename,_subgroupname="wfr",_S0=True,_S1=False,_S2=False,_S3=False,_overwrite=True):
    """
     Auxiliary function to write the Stokes parameters data into a hdf5 generic file. The Stokes parameters of a plane
     monochromatic wave are four quantities: S0, S1, S2 and S3. Only three of them are independent, since they are
     related by the identity: S0^2 = S1^2 + S2^2 + S3^2.
    :param _Stokes: input / output resulting radiation stokes Parameters (instance of SRWLStokes)
    :param _filename: path to file for saving the wavefront
    :param _subgroupname: container mechanism by which HDF5 files are organised
    :param _S0: I = P_0 + P_90 = <Ex^2 + Ey^2>
    :param _S1: Q = P_0 - P_90 = <Ex^2 - Ey^2>
    :param _S2: U = P_45 + P_135 = <2Ex*Ey*cos(delta)>
    :param _S3: V = P_r_circular + P_l_circular = <2Ex*Ey*sin(delta)>
    :param _overwrite: flag that should always be set to True to avoid infinity loop on the recursive part of the function.
    """
    try:
        if not os.path.isfile(_filename):  # if file doesn't exist, create it.
            sys.stdout.flush()
            f = h5py.File(_filename, 'w')
            # points to the default data to be plotted
            f.attrs['default']          = 'entry'
            # give the HDF5 root some more attributes
            f.attrs['file_name']        = _filename
            f.attrs['file_time']        = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
            f.attrs['creator']          = 'save_stokes_2_hdf5'
            f.attrs['code']             = 'SRW'
            f.attrs['HDF5_Version']     = h5py.version.hdf5_version
            f.attrs['h5py_version']     = h5py.version.version
            f.close()

        arxx = numpy.array(_Stokes.arS)
        arxx = arxx.reshape((4, _Stokes.mesh.ny, _Stokes.mesh.nx)).T

        if _S0:
            _dump_arr_2_hdf5(arxx[:, :, 0].T, "Stokes_S0/S0", _filename, _subgroupname)
        if _S1:
            _dump_arr_2_hdf5(arxx[:, :, 1].T, "Stokes_S1/S1", _filename, _subgroupname)
        if _S2:
            _dump_arr_2_hdf5(arxx[:, :, 2].T, "Stokes_S2/S2", _filename, _subgroupname)
        if _S3:
            _dump_arr_2_hdf5(arxx[:, :, 3].T, "Stokes_S3/S3", _filename, _subgroupname)

        f = h5py.File(_filename, 'a')
        f1 = f[_subgroupname]

        # points to the default data to be plotted
        f1.attrs['NX_class'] = 'NXentry'
        f1.attrs['default']  = 'S0'

        #f1["Stokes_method"] = "SRW"
        f1["Stokes_photon_energy"] = _Stokes.mesh.eStart
        f1["Stokes_mesh_X"] = numpy.array([_Stokes.mesh.xStart, _Stokes.mesh.xFin, _Stokes.mesh.nx])
        f1["Stokes_mesh_Y"] = numpy.array([_Stokes.mesh.yStart, _Stokes.mesh.yFin, _Stokes.mesh.ny])

        # Add NX plot attribites for automatic plot with silx view
        myflags = [_S0,_S1,_S2,_S3]
        mylabels = ['S0','S1','S2','S3']
        for i,label in enumerate(mylabels):
            if myflags[i]:

                f2 = f1['Stokes_%s'%(mylabels[i])]
                f2.attrs['NX_class'] = 'NXdata'
                f2.attrs['signal'] = '%s'%(mylabels[i])
                f2.attrs['axes'] = [b'axis_y', b'axis_x']

                # ds = nxdata.create_dataset('image_data', data=data)
                f3 = f2["%s"%(mylabels[i])]
                f3.attrs['interpretation'] = 'image'

                # X axis data
                ds = f2.create_dataset('axis_y', data=1e6*numpy.linspace(_Stokes.mesh.yStart,_Stokes.mesh.yFin,_Stokes.mesh.ny))
                # f1['axis1_name'] = numpy.arange(_wfr.mesh.ny)
                ds.attrs['units'] = 'microns'
                ds.attrs['long_name'] = 'Y Pixel Size (microns)'    # suggested X axis plot label
                #
                # Y axis data
                ds = f2.create_dataset('axis_x', data=1e6*numpy.linspace(_Stokes.mesh.xStart,_Stokes.mesh.xFin,_Stokes.mesh.nx))
                ds.attrs['units'] = 'microns'
                ds.attrs['long_name'] = 'X Pixel Size (microns)'    # suggested Y axis plot label
        f.close()

        FileName = _filename.split("/")
        print(">>>> save_stokes_2_hdf5: file witten/updated %s" %FileName[-1])

    except:
        if _overwrite is not True:
            print(">>>> Bad input argument")
            return
        os.remove(_filename)
        FileName = _filename.split("/")
        print(">>>> save_stokes_2_hdf5: file deleted %s"%FileName[-1])
        save_stokes_2_hdf5(_Stokes,_filename,_subgroupname,_S0,_S1,_S2,_S3,_overwrite = False)

def SRWdat_2_h5(_file_path,_num_type='f'):
    """
    Auxiliary to be convert output files from SRW .dat format into a generic wavefront hdf5 generic file. Read-in tabulated
    Intensity data from an ASCII file (format is defined in srwl_uti_save_intens_ascii)
    :param _file_path: path to file for saving the wavefront
    """
    file_h5 = _file_path.replace(".dat", ".h5")
    filename = file_h5.split("/")
    wfr = SRWLWfr()
    wfr.arEx, wfr.mesh = srwl_uti_read_intens_ascii(_file_path, _num_type)

    arxx = numpy.array(wfr.arEx)

    if wfr.mesh.ne > 1:
        print('>>>> No support for such .dat file')
        return
        # sys.stdout.flush()

        # f = h5py.File(file_h5, 'w')
        # Energy = numpy.linspace(wfr.mesh.eStart, wfr.mesh.eFin, num=wfr.mesh.ne)
        # arxx = arxx.reshape((wfr.mesh.ny, wfr.mesh.nx, wfr.mesh.ne, 1))
        # arxx = arxx.swapaxes(0, 2)

        # for count in range(wfr.mesh.ne):

        #     subgroupname = "wfr_%d"%count
        #     f1 = f.create_group(subgroupname)
        #     f1["converted_array"] = arxx
        #     f1["wfr_method"] = "SRW"
        #     f1["wfr_photon_energy"] = Energy[count]
        #     f1["wfr_radii"] = numpy.array([wfr.Rx, wfr.dRx, wfr.Ry, wfr.dRy])
        #     f1["wfr_mesh"] = numpy.array([wfr.mesh.xStart, wfr.mesh.xFin, wfr.mesh.nx, wfr.mesh.yStart, wfr.mesh.yFin, wfr.mesh.ny])
        # f.close()
        # return

    arxx = arxx.reshape((wfr.mesh.ny, wfr.mesh.nx))#.T

    sys.stdout.flush()

    f = h5py.File(file_h5, 'w')
    # points to the default data to be plotted
    f.attrs['default'] = 'entry'
    # give the HDF5 root some more attributes
    f.attrs['file_name'] = filename[-1]
    f.attrs['file_time'] = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
    f.attrs['creator'] = 'save_wfr_2_hdf5'
    f.attrs['code'] = 'SRW'
    f.attrs['HDF5_Version'] = h5py.version.hdf5_version
    f.attrs['h5py_version'] = h5py.version.version

    f1 = f.create_group("wfr")
    # f1["converted_array"] = arxx
    # f1["wfr_method"] = "SRW"

    fdata = f1.create_dataset("converted_array/array", data=arxx)

    f1["wfr_photon_energy"] = float(wfr.mesh.eStart)
    f1["wfr_zStart"] = wfr.mesh.zStart
    f1["wfr_Rx_dRx"] = numpy.array([wfr.Rx, wfr.dRx])
    f1["wfr_Ry_dRy"] = numpy.array([wfr.Ry, wfr.dRy])
    f1["wfr_mesh_X"] = numpy.array([wfr.mesh.xStart, wfr.mesh.xFin, wfr.mesh.nx])
    f1["wfr_mesh_Y"] = numpy.array([wfr.mesh.yStart, wfr.mesh.yFin, wfr.mesh.ny])

    f2 = f1["converted_array"]
    f2.attrs['NX_class'] = 'NXdata'
    f2.attrs['signal'] = 'array'
    f2.attrs['axes'] = [b'axis_y', b'axis_x']

    # ds = nxdata.create_dataset('image_data', data=data)
    f3 = f2['array']
    f3.attrs['interpretation'] = 'image'

    # X axis data
    ds = f2.create_dataset('axis_y', data=1e6 * numpy.linspace(wfr.mesh.yStart, wfr.mesh.yFin, wfr.mesh.ny))
    # f1['axis1_name'] = numpy.arange(_wfr.mesh.ny)
    ds.attrs['units'] = 'microns'
    ds.attrs['long_name'] = 'Y Pixel Size (microns)'  # suggested X axis plot label
    #
    # Y axis data
    ds = f2.create_dataset('axis_x', data=1e6 * numpy.linspace(wfr.mesh.xStart, wfr.mesh.xFin, wfr.mesh.nx))
    ds.attrs['units'] = 'microns'
    ds.attrs['long_name'] = 'X Pixel Size (microns)'  # suggested Y axis plot label

    f.close()


def load_hdf5_2_dictionary(filename, filepath):

    try:
        f = h5py.File(filename, 'r')

        out =  {
            "wfr_complex_amplitude_s": f[filepath+"/wfr_complex_amplitude_s"][()].T,
            "wfr_complex_amplitude_p": f[filepath+"/wfr_complex_amplitude_p"][()].T,
            "wfr_photon_energy":       f[filepath+"/wfr_photon_energy"][()],
            "wfr_zStart":              f[filepath+"/wfr_zStart"][()],
            "wfr_mesh_X":              f[filepath+"/wfr_mesh_X"][()],
            "wfr_mesh_Y":              f[filepath+"/wfr_mesh_Y"][()],
            "wfr_mesh_E":              f[filepath+"/wfr_mesh_E"][()],
            "wfr_Rx_dRx":              f[filepath+"/wfr_Rx_dRx"][()],
            "wfr_Ry_dRy":              f[filepath+"/wfr_Ry_dRy"][()],

            "wfr_typeE" :        f[filepath+"/wfr_typeE"][()],
            "wfr_xc" :           f[filepath+"/wfr_xc"][()],
            "wfr_yc" :           f[filepath+"/wfr_yc"][()],
            "wfr_avgPhotEn" :    f[filepath+"/wfr_avgPhotEn"][()],
            "wfr_presCA" :       f[filepath+"/wfr_presCA"][()],
            "wfr_presFT" :       f[filepath+"/wfr_presFT"][()],
            "wfr_unitElFld" :    f[filepath+"/wfr_unitElFld"][()],
            "wfr_unitElFldAng" : f[filepath+"/wfr_unitElFldAng"][()],

            "wfr_partBeam_Iavg":           f[filepath+"/partBeam/Iavg"][()],
            "wfr_partBeam_nPart":          f[filepath + "/partBeam/nPart"][()],
            "wfr_partBeam_particle_x":     f[filepath + "/partBeam/particle_x"][()],
            "wfr_partBeam_particle_y":     f[filepath + "/partBeam/particle_y"][()],
            "wfr_partBeam_particle_z":     f[filepath + "/partBeam/particle_z"][()],
            "wfr_partBeam_particle_xp":    f[filepath + "/partBeam/particle_xp"][()],
            "wfr_partBeam_particle_yp":    f[filepath + "/partBeam/particle_yp"][()],
            "wfr_partBeam_particle_gamma": f[filepath + "/partBeam/particle_gamma"][()],
            "wfr_partBeam_particle_relE0": f[filepath + "/partBeam/particle_relE0"][()],
            "wfr_partBeam_particle_nq":    f[filepath + "/partBeam/particle_nq"][()],
            "wfr_partBeam_arStatMom2":     f[filepath + "/partBeam/arStatMom2"][()],

            "wfr_arElecPropMatr" : f[filepath+"/wfr_arElecPropMatr"][()],
            "wfr_arMomX" :         f[filepath+"/wfr_arMomX"][()],
            "wfr_arMomY" :         f[filepath+"/wfr_arMomY"][()],
            "wfr_arWfrAuxData" :   f[filepath+"/wfr_arWfrAuxData"][()],
        }

        f.close()

        return out
    except:
        raise Exception("Failed to load SRW wavefront from h5 file: "+filename)

#
# Auxiliar functions
#
def _numpyArrayToSRWArray(numpy_array):
    """
    Converts a numpy.array to an array usable by SRW.
    :param numpy_array: a 2D numpy array
    :return: a 2D complex SRW array
    """
    elements_size = numpy_array.size

    r_horizontal_field = numpy_array[:, :].real.transpose().flatten().astype(numpy.float32)
    i_horizontal_field = numpy_array[:, :].imag.transpose().flatten().astype(numpy.float32)

    tmp = numpy.zeros(elements_size * 2, dtype=numpy.float32)
    for i in range(elements_size):
        tmp[2*i] = r_horizontal_field[i]
        tmp[2*i+1] = i_horizontal_field[i]

    return array('f', tmp)

def _SRWArrayToNumpy(srw_array, dim_x, dim_y, number_energies):
    """
    Converts a SRW array to a numpy.array.
    :param srw_array: SRW array
    :param dim_x: size of horizontal dimension
    :param dim_y: size of vertical dimension
    :param number_energies: Size of energy dimension
    :return: 4D numpy array: [energy, horizontal, vertical, polarisation={0:horizontal, 1: vertical}]
    """
    re = numpy.array(srw_array[::2],  dtype=numpy.float32)
    im = numpy.array(srw_array[1::2], dtype=numpy.float32)

    e = re + 1j * im
    e = e.reshape((dim_y,
                   dim_x,
                   number_energies,
                   1)
                  )

    e = e.swapaxes(0, 2)

    return e.copy()

def _dump_arr_2_hdf5(_arr,_calculation, _filename, _subgroupname):
    """
    Auxiliary routine to save_wfr_2_hdf5() and save_stokes_2_hdf5()
    :param _arr: (usually 2D) array to be saved on the hdf5 file inside the _subgroupname
    :param _calculation
    :param _filename: path to file for saving the wavefront
    :param _subgroupname: container mechanism by which HDF5 files are organised
    """
    sys.stdout.flush()
    f = h5py.File(_filename, 'a')
    try:
        f1 = f.create_group(_subgroupname)
    except:
        f1 = f[_subgroupname]
    #f1[_calculation] = _arr
    fdata = f1.create_dataset(_calculation, data=_arr)
    f.close()


def _dictionary_to_wfr(wdic):
    """
    Creates a SRWWavefront from pi and sigma components of the electrical field.

    """

    w_s = wdic["wfr_complex_amplitude_s"]
    w_p = wdic["wfr_complex_amplitude_p"]
    energy = wdic["wfr_photon_energy"]
    X = wdic["wfr_mesh_X"]
    Y = wdic["wfr_mesh_Y"]
    E = wdic["wfr_mesh_E"]
    RX = wdic["wfr_Rx_dRx"]
    RY = wdic["wfr_Ry_dRy"]
    Z = wdic["wfr_zStart"]

    # wshape = w_s.shape
    # print(">>>>>>>>>>>>>>>>wshape before: ",wshape,w_s.shape)
    # w_s.shape = [1,wshape[0],wshape[1],1]
    # w_p.shape = [1,wshape[0],wshape[1],1]
    # print(">>>>>>>>>>>>>>>>wshape after: ",wshape,w_s.shape)

    horizontal_size = w_s.shape[0]
    vertical_size = w_s.shape[1]

    if horizontal_size % 2 == 1 or \
       vertical_size % 2 == 1:
        # raise Exception("Both horizontal and vertical grid must have even number of points")
        print("NumpyToSRW: WARNING: Both horizontal and vertical grid must have even number of points")

    horizontal_field = _numpyArrayToSRWArray(w_s)
    vertical_field = _numpyArrayToSRWArray(w_p)

    partBeam = SRWLPartBeam(_Iavg=wdic["wfr_partBeam_Iavg"],
                            _nPart=wdic["wfr_partBeam_nPart"],
                            _partStatMom1=SRWLParticle(_x=wdic["wfr_partBeam_particle_x"],
                                                       _y=wdic["wfr_partBeam_particle_y"],
                                                       _z=wdic["wfr_partBeam_particle_z"],
                                                       _xp=wdic["wfr_partBeam_particle_xp"],
                                                       _yp=wdic["wfr_partBeam_particle_yp"],
                                                       _gamma=wdic["wfr_partBeam_particle_gamma"],
                                                       _relE0=wdic["wfr_partBeam_particle_relE0"],
                                                       _nq=wdic["wfr_partBeam_particle_nq"]),
                            _arStatMom2=array('d', wdic["wfr_partBeam_arStatMom2"]))

    srw_wavefront = SRWLWfr(_arEx=horizontal_field,
                            _arEy=vertical_field,
                            _typeE=wdic["wfr_typeE"],
                            _eStart=E[0],
                            _eFin=E[1],
                            _ne=int(E[2]),
                            _xStart=X[0],
                            _xFin=X[1],
                            _nx=int(X[2]),
                            _yStart=Y[0],
                            _yFin=Y[1],
                            _ny=int(Y[2]),
                            _zStart=Z,
                            _partBeam=partBeam)

    srw_wavefront.xc = wdic["wfr_xc"]
    srw_wavefront.yc = wdic["wfr_yc"]
    srw_wavefront.avgPhotEn = wdic["wfr_avgPhotEn"]
    srw_wavefront.presCA = wdic["wfr_presCA"]
    srw_wavefront.presFT = wdic["wfr_presFT"]
    srw_wavefront.unitElFld = wdic["wfr_unitElFld"]
    srw_wavefront.unitElFldAng = wdic["wfr_unitElFldAng"]

    srw_wavefront.Rx = RX[0]
    srw_wavefront.Ry = RY[0]
    srw_wavefront.dRx = RX[1]
    srw_wavefront.dRy = RY[1]

    srw_wavefront.arElecPropMatr = array('d', wdic["wfr_arElecPropMatr"])
    srw_wavefront.arMomX         = array('d', wdic["wfr_arMomX"])
    srw_wavefront.arMomY         = array('d', wdic["wfr_arMomY"])
    srw_wavefront.arWfrAuxData   = array('d', wdic["wfr_arWfrAuxData"])
    return srw_wavefront



SRW_INSTALLED = True
try:
    try:
        from srwpy.srwlib import *   # OASYS 2.0 - SRW 4
        from srwpy.uti_plot import *
    except:
        from srwlib import *     # SRW 3
        from uti_plot import *
except:
    try:
        from oasys_srw.srwlib import *   # OASYS 1.X
        from oasys_srw.uti_plot import *
    except:
        SRW_INSTALLED = False


import numpy

from wofrysrw.beamline.optical_elements.srw_optical_element import SRWOpticalElement
from wofrysrw.util.srw_util import numpyArraysToSRWArray, SRWArrayToNumpyArrays

from wofrysrw.util.srw import SRWLOptT

class SRWTransmission(SRWOpticalElement):
    def __init__(self,
                 x_range=None,
                 y_range=None,
                 transmission_amplitudes=None,
                 transmission_optical_path_difference=None,
                 energy=15000):
        SRWOpticalElement.__init__(self)

        self.energy                               = energy
        self.x_range                              = x_range
        self.y_range                              = y_range
        self.transmission_amplitudes              = transmission_amplitudes
        self.transmission_optical_path_difference = transmission_optical_path_difference

    def toSRWLOpt(self):
        nx = self.transmission_amplitudes.shape[0]
        ny = self.transmission_amplitudes.shape[1]
        transmission_array = numpyArraysToSRWArray(numpy_array_re=self.transmission_amplitudes,
                                                   numpy_array_im=self.transmission_optical_path_difference,
                                                   type='d')

        return SRWLOptT(_nx=nx,
                        _ny=ny,
                        _x=0.5*(self.x_range[1]+self.x_range[0]),
                        _y=0.5*(self.y_range[1]+self.y_range[0]),
                        _rx=self.x_range[1] - self.x_range[0],
                        _ry=self.y_range[1] - self.y_range[0],
                        _arTr=transmission_array,
                        _extTr=1,
                        _ne=1,
                        _eStart=self.energy,
                        _eFin=self.energy)

    def fromSRWLOpt(self, srwlopt=SRWLOptT()):
        if not isinstance(srwlopt, SRWLOptT):
            raise ValueError("SRW object is not a SRWLOptT object")

        energy  = 0.5*(srwlopt.mesh.eFin + srwlopt.mesh.eStart)
        x_range = [srwlopt.mesh.xStart, srwlopt.mesh.xFin]
        y_range = [srwlopt.mesh.yStart, srwlopt.mesh.yFin]
        transmission_amplitudes, transmission_optical_path_difference = SRWArrayToNumpyArrays(srw_array=srwlopt.arTr,
                                                                                              dim_x=srwlopt.mesh.nx,
                                                                                              dim_y=srwlopt.mesh.ny,
                                                                                              number_energies=1,
                                                                                              polarized=False)
        self.__init__(x_range=x_range,
                      y_range=y_range,
                      energy=energy,
                      transmission_amplitudes=transmission_amplitudes,
                      transmission_optical_path_difference=transmission_optical_path_difference)

    def to_python_code(self, data=None):
        oe_name = data[0]

        nx = self.transmission_amplitudes.shape[0]
        ny = self.transmission_amplitudes.shape[1]
        transmission_array = numpyArraysToSRWArray(numpy_array_re=self.transmission_amplitudes,
                                                   numpy_array_im=self.transmission_optical_path_difference)

        def to_string(srw_array):
            text = "["
            for element in srw_array: text += str(element) + ", "
            return text[:-2] + "]"

        text_code = "transmission_array = array('d', " + to_string(transmission_array) + ")\n\n"
        text_code += oe_name + "=" + "SRWLOptT(_nx=" + str(nx) + "," + "\n"
        text_code += "               _ny=" + str(ny) +  "," + "\n"
        text_code += "               _x=" + str(0.5*(self.x_range[1]+self.x_range[0])) + "," + "\n"
        text_code += "               _y=" + str(0.5*(self.y_range[1]+self.y_range[0])) + "," + "\n"
        text_code += "               _rx=" + str(self.x_range[1]-self.x_range[0]) + "," + "\n"
        text_code += "               _ry=" + str(self.y_range[1]-self.y_range[0]) + "," + "\n"
        text_code += "               _arTr=transmission_array,\n"
        text_code += "               _extTr=1,\n"
        text_code += "               _ne=1,\n"
        text_code += "               _eStart=" + str(self.energy) + "," + "\n"
        text_code += "               _eFin=" + str(self.energy) + ")" + "\n"

        return text_code



from wofrysrw.util.srw import SRWLMagFldC, SRWLMagFld3D, array

from syned.storage_ring.magnetic_structure import MagneticStructure
from wofrysrw.storage_ring.srw_magnetic_structure import SRWMagneticStructure

# from original SRW Example 01, by Oleg Chubar (BNL)
def AuxReadInMagFld3D(filePath, sCom):
    f = open(filePath, 'r')
    f.readline()  # 1st line: just pass

    xStart = float(f.readline().split(sCom, 2)[1])  # 2nd line: initial X position [m]; it will not actually be used
    xStep = float(f.readline().split(sCom, 2)[1])  # 3rd line: step vs X [m]
    xNp = int(f.readline().split(sCom, 2)[1])  # 4th line: number of points vs X
    yStart = float(f.readline().split(sCom, 2)[1])  # 5th line: initial Y position [m]; it will not actually be used
    yStep = float(f.readline().split(sCom, 2)[1])  # 6th line: step vs Y [m]
    yNp = int(f.readline().split(sCom, 2)[1])  # 7th line: number of points vs Y
    zStart = float(f.readline().split(sCom, 2)[1])  # 8th line: initial Z position [m]; it will not actually be used
    zStep = float(f.readline().split(sCom, 2)[1])  # 9th line: step vs Z [m]
    zNp = int(f.readline().split(sCom, 2)[1])  # 10th line: number of points vs Z

    totNp = xNp * yNp * zNp

    locArBx = array('d', [0] * totNp)
    locArBy = array('d', [0] * totNp)
    locArBz = array('d', [0] * totNp)

    for i in range(totNp):
        curLineParts = f.readline().split('\t')
        if len(curLineParts) == 3:
            locArBx[i] = float(curLineParts[0].strip())
            locArBy[i] = float(curLineParts[1].strip())
            locArBz[i] = float(curLineParts[2].strip())

    f.close()

    xRange = xStep
    if xNp > 1: xRange = (xNp - 1) * xStep
    yRange = yStep
    if yNp > 1: yRange = (yNp - 1) * yStep
    zRange = zStep
    if zNp > 1: zRange = (zNp - 1) * zStep

    return SRWLMagFld3D(locArBx, locArBy, locArBz, xNp, yNp, zNp, xRange, yRange, zRange, 1)

class SRW3DMagneticStructure(MagneticStructure, SRWMagneticStructure):

    def __init__(self,
                 file_name="",
                 comment_character="#",
                 interpolation_method=1):
        MagneticStructure.__init__(self)

        self.file_name = file_name
        self.comment_character = comment_character
        self.interpolation_method=interpolation_method

    def get_SRWMagneticStructure(self):
        return AuxReadInMagFld3D(self.file_name, self.comment_character)

    def get_SRWLMagFldC(self):
        magnetic_field_container = SRWLMagFldC()  # Container
        magnetic_field_container.allocate(1)  # Magnetic Field consists of 1 part

        magnetic_field_container.arMagFld[0] = self.get_SRWMagneticStructure()
        magnetic_field_container.arMagFld[0].interp = self.interpolation_method
        magnetic_field_container.arXc[0] = self.horizontal_central_position
        magnetic_field_container.arYc[0] = self.vertical_central_position
        magnetic_field_container.arZc[0] = self.longitudinal_central_position

        magnetic_field_container.arMagFld[0].nRep = 1

        return magnetic_field_container

    @classmethod
    def get_default_initial_z(cls, file_name, comment_character="#", longitudinal_central_position=0):
        return longitudinal_central_position - 0.5*AuxReadInMagFld3D(file_name, comment_character).rz

    @classmethod
    def get_source_length(cls, file_name, comment_character="#"):
        return AuxReadInMagFld3D(file_name, comment_character).rz

    def to_python_code(self, data=None):
        text_code  = self.to_python_code_aux()
        text_code += "magnetic_field_container = SRWLMagFldC()" + "\n"
        text_code += "magnetic_field_container.allocate(1)" + "\n"
        text_code += "magnetic_field_container.arMagFld[0] = magnetic_structure" + "\n"
        text_code += "magnetic_field_container.arMagFld[0].interp = " + str(self.interpolation_method) + "\n"
        text_code += "magnetic_field_container.arXc[0] = " + str(self.horizontal_central_position) + "\n"
        text_code += "magnetic_field_container.arYc[0] = " + str(self.vertical_central_position) + "\n"
        text_code += "magnetic_field_container.arZc[0] = " + str(self.longitudinal_central_position) + "\n"

        return text_code

    def to_python_code_aux(self):
        text_code = "def AuxReadInMagFld3D(filePath, sCom):" + "\n" + \
                    "    f = open(filePath, 'r')" + "\n" + \
                    "    f.readline()  # 1st line: just pass" + "\n\n" + \
                    "    xStart = float(f.readline().split(sCom, 2)[1])  # 2nd line: initial X position [m]; it will not actually be used" + "\n" + \
                    "    xStep  = float(f.readline().split(sCom, 2)[1])  # 3rd line: step vs X [m]" + "\n" + \
                    "    xNp = int(f.readline().split(sCom, 2)[1])  # 4th line: number of points vs X" + "\n" + \
                    "    yStart = float(f.readline().split(sCom, 2)[1])  # 5th line: initial Y position [m]; it will not actually be used" + "\n" + \
                    "    yStep = float(f.readline().split(sCom, 2)[1])  # 6th line: step vs Y [m]" + "\n" + \
                    "    yNp = int(f.readline().split(sCom, 2)[1])  # 7th line: number of points vs Y" + "\n" + \
                    "    zStart = float(f.readline().split(sCom, 2)[1])  # 8th line: initial Z position [m]; it will not actually be used" + "\n" + \
                    "    zStep = float(f.readline().split(sCom, 2)[1])  # 9th line: step vs Z [m]" + "\n" + \
                    "    zNp = int(f.readline().split(sCom, 2)[1])  # 10th line: number of points vs Z" + "\n" + \
                    "    totNp = xNp * yNp * zNp" + "\n" + \
                    "    locArBx = array('d', [0] * totNp)" + "\n" + \
                    "    locArBy = array('d', [0] * totNp)" + "\n" + \
                    "    locArBz = array('d', [0] * totNp)" + "\n\n" + \
                    "    for i in range(totNp):" + "\n" + \
                    "        curLineParts = f.readline().split('\t')" + "\n" + \
                    "        if len(curLineParts) == 3:" + "\n" + \
                    "            locArBx[i] = float(curLineParts[0].strip())" + "\n" + \
                    "            locArBy[i] = float(curLineParts[1].strip())" + "\n" + \
                    "            locArBz[i] = float(curLineParts[2].strip())" + "\n" + \
                    "    f.close()" + "\n" + \
                    "    xRange = xStep" + "\n" + \
                    "    if xNp > 1: xRange = (xNp - 1) * xStep" + "\n" + \
                    "    yRange = yStep" + "\n" + \
                    "    if yNp > 1: yRange = (yNp - 1) * yStep" + "\n" + \
                    "    zRange = zStep" + "\n" + \
                    "    if zNp > 1: zRange = (zNp - 1) * zStep" + "\n" + \
                    "    return SRWLMagFld3D(locArBx, locArBy, locArBz, xNp, yNp, zNp, xRange, yRange, zRange, 1)\n\n"

        text_code += "magnetic_structure = AuxReadInMagFld3D(\"" + str(self.file_name) + "\",\"" + str(self.comment_character) + "\")" + "\n"

        return text_code




import numpy

from wofrysrw.storage_ring.srw_light_source import SRWLightSource
from wofrysrw.storage_ring.light_sources.srw_undulator_light_source import SRWUndulatorLightSource
from wofrysrw.storage_ring.srw_light_source import SourceWavefrontParameters

from srxraylib.plot.gol import plot, plot_contour, plot_surface

period_length=0.02
harmonic = 7

undulator = SRWUndulatorLightSource("MicroXRD",
                                    electron_energy_in_GeV=2.0,
                                    electron_energy_spread=0.0007,
                                    ring_current=0.4,
                                    electron_beam_size_h=0.05545e-3,
                                    electron_beam_size_v=2.784e-6,
                                    emittance=0.2525e-9,
                                    coupling_costant=0.01,
                                    K_vertical=1.5,
                                    period_length=period_length,
                                    number_of_periods=int(1.5/period_length))

print(undulator.get_electron_beam().get_electron_beam_geometrical_properties().to_info())

resonance_energy = undulator.get_resonance_energy()

print("Harmonic Energy", resonance_energy*harmonic)

properties = undulator.get_photon_source_properties(harmonic=harmonic)

print(properties.to_info())

wf_parameters = SourceWavefrontParameters(photon_energy_min = resonance_energy*harmonic,
                                          photon_energy_max = resonance_energy*harmonic,
                                          photon_energy_points=1,
                                          h_slit_gap = 0.005,
                                          v_slit_gap = 0.005,
                                          h_slit_points=51,
                                          v_slit_points=51,
                                          distance = 10.0)

e, h, v, i = undulator.get_intensity(source_wavefront_parameters=wf_parameters)

plot_contour(i[int(e.size/2)],h*1e3,v*1e3,title="%s SRW; E=%g eV"%("MicroXRD",e[int(e.size/2)]),xtitle="H [mm]",ytitle="V [mm]",plot_points=0,
             contour_levels=numpy.linspace(0, i.max(), 20), cmap=None, cbar=1, cbar_title="Intensity [ph/s/.1%bw/mm^2]",show=1)

plot_surface(i[int(e.size/2)],h*1e3,v*1e3,title="%s SRW; E=%g eV"%("MicroXRD",e[int(e.size/2)]),xtitle="H [mm]",ytitle="V [mm]", show=1)

# ------------------------------------------
'''
wf_parameters = SourceWavefrontParameters(photon_energy_min = 1,
                                          photon_energy_max = 12001,
                                          photon_energy_points=12000,
                                          h_slit_gap = 0.001,
                                          v_slit_gap = 0.001,
                                          h_slit_points=10,
                                          v_slit_points=50,
                                          distance = 10.0)

e, i = undulator.get_spectral_flux(source_wavefront_parameters=wf_parameters, multi_electron=True)

plot(e, i, show=1, title="Flux for MicroXRD", xtitle="Energy [eV]",  ytitle="Flux [ph/s/.1%bw]")
'''

wf_parameters = SourceWavefrontParameters(photon_energy_min = 1,
                                          photon_energy_max = 12001,
                                          photon_energy_points=12000,
                                          h_slit_gap = 0.001,
                                          v_slit_gap = 0.001,
                                          h_slit_points=1,
                                          v_slit_points=1,
                                          distance = 10.0)

e, i = undulator.get_undulator_flux(source_wavefront_parameters=wf_parameters)

plot(e, i, show=1, title="Flux for MicroXRD", xtitle="Energy [eV]",  ytitle="Flux [ph/s/.1%bw]")

# ------------------------------------------

wf_parameters = SourceWavefrontParameters(h_slit_gap = 0.005,
                                          v_slit_gap = 0.005,
                                          h_slit_points=51,
                                          v_slit_points=51,
                                          distance = 10.0)

h, v, p = undulator.get_power_density(source_wavefront_parameters=wf_parameters)

total_power = SRWLightSource.get_total_power_from_power_density(h, v, p)

plot_contour(p, h*1e3, v*1e3,title="%s SRW, total power = %g"%("Micro XRD", total_power),xtitle="H [mm]",ytitle="V [mm]",plot_points=0,
             contour_levels=numpy.linspace(0,numpy.max([p.max()]), 100), cmap=None,cbar=1, cbar_title="Power density [$W/mm^2$]",show=1)

plot_surface(p,h*1e3,v*1e3,  title="%s SRW, total power = %g"%("Micro XRD", total_power),xtitle="H [mm]",ytitle="V [mm]",show=1)





from syned.storage_ring.magnetic_structure import MagneticStructure
from wofrysrw.util.srw import array, SRWLMagFldC

from wofrysrw.srw_object import SRWObject

class SRWMagneticStructureDecorator():

    def get_SRWMagneticStructure(self):
        raise NotImplementedError("this method should be implented in subclasses")

    def get_SRWLMagFldC(self):
        return SRWLMagFldC(_arMagFld=[self.get_SRWMagneticStructure()],
                           _arXc=array('d', [self.horizontal_central_position]),
                           _arYc=array('d', [self.vertical_central_position]),
                           _arZc=array('d', [self.longitudinal_central_position]))


class SRWMagneticStructure(SRWMagneticStructureDecorator, SRWObject):
    def __init__(self,
                 horizontal_central_position = 0.0,
                 vertical_central_position = 0.0,
                 longitudinal_central_position = 0.0):
        super().__init__()

        self.horizontal_central_position = horizontal_central_position
        self.vertical_central_position = vertical_central_position
        self.longitudinal_central_position = longitudinal_central_position

    def to_python_code(self, data=None):
        text_code  = self.to_python_code_aux()
        text_code += "magnetic_field_container = SRWLMagFldC(_arMagFld=[magnetic_structure], " + "\n"
        text_code += "                                       _arXc=array('d', [" + str(self.horizontal_central_position) + "]), " + "\n"
        text_code += "                                       _arYc=array('d', [" + str(self.vertical_central_position) + "]), " + "\n"
        text_code += "                                       _arZc=array('d', [" + str(self.longitudinal_central_position) + "]))" + "\n"

        return text_code

    def to_python_code_aux(self):
        raise NotImplementedError()



#! /usr/bin/env python3

import os

try:
    from setuptools import find_packages, setup
except AttributeError:
    from setuptools import find_packages, setup

NAME = 'wofrysrw'
VERSION = '1.1.34'
ISRELEASED = True

DESCRIPTION = 'WOFRY for SRW library'
README_FILE = os.path.join(os.path.dirname(__file__), 'README.md')
LONG_DESCRIPTION = open(README_FILE).read()
AUTHOR = 'Luca Rebuffi'
AUTHOR_EMAIL = 'lrebuffi@anl.gov'
URL = 'https://github.com/lucarebuffi/wofrysrw'
DOWNLOAD_URL = 'https://github.com/lucarebuffi/wofrysrw'
LICENSE = 'GPLv3'

KEYWORDS = [
    'dictionary',
    'glossary',
    'synchrotron'
    'simulation',
]

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: X11 Applications :: Qt',
    'Environment :: Console',
    'Environment :: Plugins',
    'Programming Language :: Python :: 3',
    'Intended Audience :: Science/Research',
]

SETUP_REQUIRES = (
    'setuptools',
)

INSTALL_REQUIRES = (
    'setuptools',
    'wofry>=1.0.31'
)

PACKAGES = ["wofrysrw",]
PACKAGE_DATA = {}

if __name__ == '__main__':
    setup(
          name = NAME,
          version = VERSION,
          description = DESCRIPTION,
          long_description = LONG_DESCRIPTION,
          author = AUTHOR,
          author_email = AUTHOR_EMAIL,
          url = URL,
          download_url = DOWNLOAD_URL,
          license = LICENSE,
          keywords = KEYWORDS,
          classifiers = CLASSIFIERS,
          packages = PACKAGES,
          package_data = PACKAGE_DATA,
          setup_requires = SETUP_REQUIRES,
          install_requires = INSTALL_REQUIRES,
          include_package_data = True,
          zip_safe = False,
        )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["R Celestre, M Sanchez del Rio"]
__license__ = "MIT"
__date__ = "27/05/2018"

"""

Example of dumping and loading SRW wavefronts to hdf5 files

"""


import numpy
from wofrysrw.util.srw import *

from wofrysrw.util.srw_hdf5 import save_wfr_2_hdf5, load_hdf5_2_wfr, load_hdf5_2_dictionary


import scipy.constants as codata

try:
    import matplotlib.pylab as plt
    # plt.switch_backend("Qt5Agg")
    from srxraylib.plot.gol import plot_image
except:
    raise Exception("Failed to import graphics tools.")


#
# main code
#

def plot_wfr(wfr,kind='intensity',show=True,xtitle="X",ytitle="Y",title="",aspect='auto'):

	if kind == 'intensity':
		ar1 = array('f', [0]*wfr.mesh.nx*wfr.mesh.ny) # "flat" 2D array to take intensity data
		srwl.CalcIntFromElecField(ar1, wfr, 6, 0, 3, wfr.mesh.eStart, 0, 0)
	elif kind == 'phase':
		ar1 = array('d', [0]*wfr.mesh.nx*wfr.mesh.ny) # "flat" array to take 2D phase data (note it should be 'd')
		srwl.CalcIntFromElecField(ar1, wfr, 0, 4, 3, wfr.mesh.eStart, 0, 0)
	else:
		raise Exception("Unknown kind of calculation: %s"%(kind))

	arxx = numpy.array(ar1)
	arxx = arxx.reshape((wfr.mesh.ny,wfr.mesh.nx)).T
	x = numpy.linspace(1e6*wfr.mesh.xStart, 1e6*wfr.mesh.xFin, wfr.mesh.nx)
	y = numpy.linspace(1e6*wfr.mesh.yStart, 1e6*wfr.mesh.yFin, wfr.mesh.ny)

	plot_image(arxx, x, y, xtitle="%s (%d pixels)"%(xtitle,x.size), ytitle="%s (%d pixels)"%(ytitle,y.size), title=title, aspect=aspect,show=show)

	return ar1,x,y

def calculate_undulator_source(Source="EBS",pMltLr=28.3,do_plots=True):
    #############################################################################
    # Photon source
    #********************************Undulator parameters
    numPer = 77			# Number of ID Periods
    undPer = 0.0183		# Period Length [m]
    phB = 0	        	# Initial Phase of the Horizontal field component
    sB = 1		        # Symmetry of the Horizontal field component vs Longitudinal position
    xcID = 0 			# Transverse Coordinates of Undulator Center [m]
    ycID = 0
    zcID = 0
    n = 1
    beamE = 17
    #********************************Storage ring parameters

    # Wavelength = 1E-10*12.39841975/beamE
    Wavelength = codata.h*codata.c/codata.e/(1e3*beamE)


    # these first order moments CONTAIN the initial condition of the electron (X,X',Y,Y') (energy comes later)
    eBeam = SRWLPartBeam()
    eBeam.Iavg = 0.2             # average Current [A]
    eBeam.partStatMom1.x = 0.
    eBeam.partStatMom1.y = 0.
    eBeam.partStatMom1.z = -0.5*undPer*(numPer + 4) # initial Longitudinal Coordinate (set before the ID)
    eBeam.partStatMom1.xp = 0.   					# initial Relative Transverse Velocities
    eBeam.partStatMom1.yp = 0.

    electron_rest_energy_in_GeV = codata.electron_mass*codata.c**2/codata.e*1e-9
    KtoBfactor = codata.e/(2*pi*codata.electron_mass*codata.c)

    #
    # obviously these emittances value (with exception of the electron_energy) are not used for
    # the single electron calculation
    #
    if (Source.lower() == 'ebs'):
        # e- beam paramters (RMS) EBS
        sigEperE = 9.3E-4 			# relative RMS energy spread
        sigX  = 30.3E-06			# horizontal RMS size of e-beam [m]
        sigXp = 4.4E-06				# horizontal RMS angular divergence [rad]
        sigY  = 3.6E-06				# vertical RMS size of e-beam [m]
        sigYp = 1.46E-06			# vertical RMS angular divergence [rad]
        electron_energy_in_GeV = 6.00
    else:
        # e- beam paramters (RMS) ESRF @ low beta
        sigEperE = 1.1E-3 			# relative RMS energy spread
        sigX     = 48.6E-06			# horizontal RMS size of e-beam [m]
        sigXp    = 106.9E-06			# horizontal RMS angular divergence [rad]
        sigY     = 3.5E-06				# vertical RMS size of e-beam [m]
        sigYp    = 1.26E-06			# vertical RMS angular divergence [rad]
        electron_energy_in_GeV = 6.04

    eBeam.partStatMom1.gamma = electron_energy_in_GeV/electron_rest_energy_in_GeV # Relative Energy
    K = sqrt(2)*sqrt(((Wavelength*2*n*eBeam.partStatMom1.gamma**2)/undPer)-1)
    print("K: ",K)
    B = K/(undPer*KtoBfactor)	# Peak Horizontal field [T] (undulator)


    # 2nd order stat. moments
    eBeam.arStatMom2[0] = sigX*sigX			 # <(x-<x>)^2>
    eBeam.arStatMom2[1] = 0					 # <(x-<x>)(x'-<x'>)>
    eBeam.arStatMom2[2] = sigXp*sigXp		 # <(x'-<x'>)^2>
    eBeam.arStatMom2[3] = sigY*sigY		     # <(y-<y>)^2>
    eBeam.arStatMom2[4] = 0					 # <(y-<y>)(y'-<y'>)>
    eBeam.arStatMom2[5] = sigYp*sigYp		 # <(y'-<y'>)^2>
    eBeam.arStatMom2[10] = sigEperE*sigEperE # <(E-<E>)^2>/<E>^2

    # Electron trajectory
    eTraj = 0

    # Precision parameters
    arPrecSR = [0]*7
    arPrecSR[0] = 1		# SR calculation method: 0- "manual", 1- "auto-undulator", 2- "auto-wiggler"
    arPrecSR[1] = 0.01	# relative precision
    arPrecSR[2] = 0		# longitudinal position to start integration (effective if < zEndInteg)
    arPrecSR[3] = 0		# longitudinal position to finish integration (effective if > zStartInteg)
    arPrecSR[4] = 20000	# Number of points for trajectory calculation
    arPrecSR[5] = 1		# Use "terminating terms"  or not (1 or 0 respectively)
    arPrecSR[6] = 0		# sampling factor for adjusting nx, ny (effective if > 0)	# -1 @Petra


    und = SRWLMagFldU([SRWLMagFldH(n, 'v', B, phB, sB, 1)], undPer, numPer)

    magFldCnt = SRWLMagFldC([und], array('d', [xcID]), array('d', [ycID]), array('d', [zcID]))

    #********************************Wavefronts
    # Monochromatic wavefront
    wfr = SRWLWfr()
    wfr.allocate(1, 512, 256)  # Photon Energy, Horizontal and Vertical Positions
    wfr.mesh.zStart = pMltLr
    wfr.mesh.eStart = beamE*1E3
    wfr.mesh.eFin = wfr.mesh.eStart
    wfr.mesh.xStart = -2.5*1E-3
    wfr.mesh.xFin = - wfr.mesh.xStart
    wfr.mesh.yStart = -1*1E-3
    wfr.mesh.yFin = - wfr.mesh.yStart
    wfr.partBeam = eBeam

    print('source calculation starts ... ')
    srwl.CalcElecFieldSR(wfr, eTraj, magFldCnt, arPrecSR)

    #
    # plot source
    #
    if do_plots:
        plot_wfr(wfr,kind='intensity',title='Source Intensity at ' + str(wfr.mesh.eStart) + ' eV',
                 xtitle='Horizontal Position [\u03bcm]',
                 ytitle='Vertical Position [\u03bcm]',aspect=None,show=True)


    print('\nsource calculation finished\n')
    return wfr

def propagate_beamline(wfr,do_plots=True):

    print("beamline calculations starts...")

    pMltLr = 28.3
    pSlt   = 40


    Drft1  = SRWLOptD((pSlt-pMltLr))

    W_MltLr = 13*1E-3
    L_MltLr = 120*1E-3
    grzAngl = 31.42*1E-3
    oeAptrMltLr = SRWLOptA('r','a',L_MltLr*numpy.sin(grzAngl),W_MltLr)

    fMltLrh = 1/((1/pMltLr)+(1/(pSlt-pMltLr)))
    fMltLrv =1E23
    oeMltLr = SRWLOptL(_Fx=fMltLrh, _Fy=fMltLrv)


    #============= Wavefront Propagation Parameters =======================#
    #                [ 0] [1] [2]  [3]  [4]  [5]  [6]  [7]  [8]  [9] [10] [11]
    ppAptrMltLr		=[ 0,  0, 1.,   0,   0,  1.,  4.,  1.,  4.,   0,   0,   0]
    ppMltLr			=[ 0,  0, 1.,   0,   0,  1.,  1.,  1.,  1.,   0,   0,   0]
    ppDrft1 		=[ 0,  0, 1.,   2,   0,  1.,  1.,  1.,  1.,   0,   0,   0]


    optBL = SRWLOptC(
                    [oeAptrMltLr,   oeMltLr,   Drft1],
                    [ppAptrMltLr,   ppMltLr, ppDrft1]
                     )

    srwl.PropagElecField(wfr, optBL)

    if do_plots:
        plot_wfr(wfr,kind='intensity',title='Focal Intensity at ' + str(wfr.mesh.eStart) + ' eV',
                 xtitle='Horizontal Position [\u03bcm]',
                 ytitle='Vertical Position [\u03bcm]',show=True)

    print('\nbeamline calculation finished\n')
    return wfr

if __name__ == "__main__":

    do_calculate = True
    do_load = True
    do_compare = True

    show_plots = False

    if do_calculate:
        wfr = calculate_undulator_source(do_plots=show_plots)

        save_wfr_2_hdf5(wfr,"tmp2.h5",intensity=True,phase=True,overwrite=True)

        wfr_end = propagate_beamline(wfr,do_plots=show_plots)

        save_wfr_2_hdf5(wfr_end,"tmp2.h5",intensity=True,phase=False,overwrite=False,subgroupname="wfr_end")

    if do_load:
        wfr_loaded = load_hdf5_2_wfr("tmp2.h5","wfr")

        save_wfr_2_hdf5(wfr_loaded,"tmp2bis.h5",intensity=True,phase=False,overwrite=True)

        wfr_end2 = propagate_beamline(wfr_loaded,do_plots=False)

        save_wfr_2_hdf5(wfr_end2,"tmp2bis.h5",intensity=True,phase=True,overwrite=False,subgroupname="wfr_end")



    if do_compare:
        wf1_source = load_hdf5_2_dictionary("tmp2.h5","wfr")
        wf2_source = load_hdf5_2_dictionary("tmp2bis.h5","wfr")
        print("comparing wavefront at source")
        for key in wf1_source.keys():
            print("   checking field: ",key)
            numpy.testing.assert_almost_equal(wf1_source[key],wf2_source[key])

        wf1_end = load_hdf5_2_dictionary("tmp2.h5","wfr_end")
        wf2_end = load_hdf5_2_dictionary("tmp2bis.h5","wfr_end")
        print("comparing wavefront propagated")
        for key in wf1_source.keys():
            print("   checking field: ",key)
            numpy.testing.assert_almost_equal(1e-6*wf1_end[key],1e-6*wf2_end[key],1)
